"""Jupyter kernel for MARPLE - flat-array APL interpreter."""

from ipykernel.kernelbase import Kernel

from marple import __version__
from marple.adapters.buffered_console import BufferedConsole
from marple.engine import Interpreter
from marple.errors import APLError
from marple.glyphs import expand_glyphs
from marple.jupyter.html_render import aplarray_to_html, ARRAY_CSS
from marple.system_commands import run_system_command


class MARPLEKernel(Kernel):
    implementation = 'marple'
    implementation_version = __version__
    language_info = {
        'name': 'apl',
        'mimetype': 'text/x-apl',
        'file_extension': '.apl',
        'codemirror_mode': 'apl',
        'pygments_lexer': 'apl',
    }
    banner = f"MARPLE v{__version__} - flat-array APL interpreter"

    def __init__(self, **kwargs: object) -> None:
        super().__init__(**kwargs)
        self._console = BufferedConsole()
        self.interp = Interpreter(io=1, console=self._console)
        self._css_sent = False

    async def do_execute(  # type: ignore[override]
        self, code: str, silent: bool, store_history: bool = True,
        user_expressions: object = None, allow_stdin: bool = False,
    ) -> dict[str, object]:
        if not self._css_sent and not silent:
            self.send_response(self.iopub_socket, 'display_data', {
                'data': {'text/html': ARRAY_CSS},
                'metadata': {},
            })
            self._css_sent = True

        code = expand_glyphs(code.strip())
        if not code:
            return self._ok_reply()

        if code.startswith(')'):
            return await self._handle_system_command(code, silent)

        try:
            self._console.clear()
            r = self.interp.execute(code)
            console_output = self._console.output
            if not silent and console_output:
                self.send_response(self.iopub_socket, 'stream', {
                    'name': 'stdout',
                    'text': console_output,
                })
            if not silent and not r.silent:
                html = aplarray_to_html(r.value)
                self.send_response(self.iopub_socket, 'execute_result', {
                    'execution_count': self.execution_count,
                    'data': {
                        'text/plain': r.display_text,
                        'text/html': html,
                    },
                    'metadata': {},
                })
            return self._ok_reply()
        except APLError as e:
            return self._error_reply(e.name, str(e))
        except Exception as e:
            return self._error_reply(type(e).__name__, str(e))

    async def _handle_system_command(self, cmd: str,
                                     silent: bool) -> dict[str, object]:
        output, _ = run_system_command(self.interp, cmd)
        if not silent and output:
            self.send_response(self.iopub_socket, 'stream', {
                'name': 'stdout',
                'text': output + '\n',
            })
        return self._ok_reply()

    async def do_complete(self, code: str,  # type: ignore[override]
                          cursor_pos: int) -> dict[str, object]:
        token = self._extract_token(code, cursor_pos)
        if not token:
            return self._empty_complete(cursor_pos)
        all_names = (
            [name for name, _ in self.interp.env.list_variables()] +
            [name for name, _ in self.interp.env.list_functions()]
        )
        matches = sorted(n for n in all_names if n.startswith(token))
        return {
            'matches': matches,
            'cursor_start': cursor_pos - len(token),
            'cursor_end': cursor_pos,
            'metadata': {},
            'status': 'ok',
        }

    async def do_inspect(self, code: str, cursor_pos: int,  # type: ignore[override]
                         detail_level: int = 0,
                         omit_sections: tuple[str, ...] = ()) -> dict[str, object]:
        token = self._extract_token(code, cursor_pos)
        if not token:
            return {'status': 'ok', 'found': False, 'data': {}, 'metadata': {}}
        val = self.interp.env.get(token)
        if val is None:
            return {'status': 'ok', 'found': False, 'data': {}, 'metadata': {}}
        info = self._describe(token, val)
        return {
            'status': 'ok',
            'found': True,
            'data': {'text/plain': info},
            'metadata': {},
        }

    async def do_is_complete(self, code: str) -> dict[str, object]:  # type: ignore[override]
        from marple.parser import is_balanced
        if is_balanced(code):
            return {'status': 'complete'}
        return {'status': 'incomplete', 'indent': '  '}

    def _extract_token(self, code: str, cursor_pos: int) -> str:
        i = cursor_pos - 1
        while i >= 0 and (code[i].isalnum() or code[i] in '_⎕∆⍙'):
            i -= 1
        return code[i + 1:cursor_pos]

    def _describe(self, name: str, val: object) -> str:
        from marple.arraymodel import APLArray
        lines = [f"Name: {name}"]
        if isinstance(val, APLArray):
            shape_str = ' '.join(str(s) for s in val.shape) if val.shape else 'scalar'
            lines.append(f"Shape: {shape_str}")
            lines.append(f"Rank: {len(val.shape)}")
            lines.append(f"Elements: {len(val.data)}")
        else:
            source = self.interp.env.get_source(name)
            if source:
                lines.append(f"Source: {source}")
            else:
                lines.append(f"Type: {type(val).__name__}")
        return '\n'.join(lines)

    def _ok_reply(self) -> dict[str, object]:
        return {
            'status': 'ok',
            'execution_count': self.execution_count,
            'payload': [],
            'user_expressions': {},
        }

    def _error_reply(self, ename: str, evalue: str) -> dict[str, object]:
        content: dict[str, object] = {
            'ename': ename,
            'evalue': evalue,
            'traceback': [f'\x1b[0;31m{ename}\x1b[0m: {evalue}'],
        }
        self.send_response(self.iopub_socket, 'error', content)
        return {'status': 'error', **content}

    def _empty_complete(self, cursor_pos: int) -> dict[str, object]:
        return {
            'matches': [],
            'cursor_start': cursor_pos,
            'cursor_end': cursor_pos,
            'metadata': {},
            'status': 'ok',
        }



"""TerminalConsole — real Console adapter for interactive terminal use."""

import sys

from marple.ports.console import Console


class TerminalConsole(Console):
    """Console adapter that reads from stdin and writes to stdout."""

    def __init__(self) -> None:
        self._use_terminal = False
        try:
            from marple.terminal import read_line
            if sys.stdin.isatty():
                self._terminal_read_line = read_line
                self._use_terminal = True
        except ImportError:
            pass

    def read_line(self, prompt: str) -> str | None:
        if self._use_terminal:
            return self._terminal_read_line()
        try:
            return input(prompt)
        except (EOFError, KeyboardInterrupt):
            print()
            return None

    def write(self, text: str) -> None:
        sys.stdout.write(text)
        sys.stdout.flush()

    def writeln(self, text: str) -> None:
        print(text)



"""Monadic primitive function dispatch for MARPLE."""

from __future__ import annotations

from typing import TYPE_CHECKING

from marple.arraymodel import APLArray, S

if TYPE_CHECKING:
    from marple.environment import Environment
from marple.errors import DomainError
from marple.functions import (
    negate,
    reciprocal,
    ceiling,
    floor,
    exponential,
    factorial,
    natural_log,
    absolute_value,
    logical_not,
    pi_times,
)
import random as _random

from marple.structural import (
    grade_down,
    grade_up,
    ravel,
    reverse,
    reverse_first,
    shape,
    transpose,
    matrix_inverse,
)


class MonadicFunctionBinding:
    """Dispatches and applies monadic primitive functions."""

    _SIMPLE: dict[str, object] = {
        "+": lambda omega: omega,
        "-": negate,
        "×": lambda omega: S((-1 if omega.data[0] < 0 else 1 if omega.data[0] > 0 else 0)),
        "÷": reciprocal,
        "⌈": ceiling,
        "⌊": floor,
        "*": exponential,
        "⍟": natural_log,
        "|": absolute_value,
        "~": logical_not,
        "⍴": shape,
        ",": ravel,
        "⌽": reverse,
        "⊖": reverse_first,
        "⍉": transpose,
        "⌹": matrix_inverse,
        "○": pi_times,
        "!": factorial,
    }

    _ENV_DEPENDENT: dict[str, str] = {
        "⍳": "_iota",
        "≢": "_tally",
        "⍋": "_grade_up",
        "⍒": "_grade_down",
        "?": "_roll",
        "⍕": "_format",
    }

    def __init__(self, env: Environment) -> None:
        self._env = env

    def apply(self, glyph: str, operand: APLArray) -> APLArray:
        """Apply a monadic primitive function to an operand."""
        method_name = self._ENV_DEPENDENT.get(glyph)
        if method_name is not None:
            return getattr(self, method_name)(operand)
        func = self._SIMPLE.get(glyph)
        if func is not None:
            return func(operand)  # type: ignore[operator]
        raise DomainError(f"Unknown monadic function: {glyph}")

    def _iota(self, operand: APLArray) -> APLArray:
        io = self._env.io
        n = int(operand.data[0])
        return APLArray([n], list(range(io, n + io)))

    def _tally(self, operand: APLArray) -> APLArray:
        return S(1) if operand.is_scalar() else S(operand.shape[0])

    def _grade_up(self, operand: APLArray) -> APLArray:
        return grade_up(operand, self._env.io)

    def _grade_down(self, operand: APLArray) -> APLArray:
        return grade_down(operand, self._env.io)

    def _roll(self, operand: APLArray) -> APLArray:
        """Monadic ?: roll. ?N → random int ⎕IO..N, ?0 → random float [0,1)."""
        io = self._env.io
        def roll_one(v: object) -> object:
            n = int(v)  # type: ignore[arg-type]
            return _random.random() if n == 0 else _random.randint(io, n - 1 + io)
        if operand.is_scalar():
            return S(roll_one(operand.data[0]))
        return APLArray(list(operand.shape), [roll_one(v) for v in operand.data])

    def _format(self, operand: APLArray) -> APLArray:
        from marple.formatting import format_num
        if operand.is_scalar():
            s = format_num(operand.data[0])
        else:
            parts = [format_num(val) for val in operand.data]
            s = " ".join(parts)
        return APLArray([len(s)], list(s))



"""Class-based APL interpreter for MARPLE."""

from marple.arraymodel import APLArray, S
from marple.backend import (
    _DOWNCAST_CT, is_numeric_array, maybe_downcast,
)
from marple.dfn_binding import DfnBinding
from marple.environment import Environment
from marple.formatting import format_result
from marple.ports.console import Console
from marple.ports.filesystem import FileSystem
from marple.executor import Executor, _newlines_to_diamonds
from marple.parser import Assignment, Program, parse
from marple.symbol_table import NC_FUNCTION, NC_OPERATOR


class EvalResult:
    """Result of evaluating an APL expression."""

    def __init__(self, value: APLArray, silent: bool, display_text: str) -> None:
        self.value = value
        self.silent = silent
        self.display_text = display_text


_SYS_FUNCTION_NAMES = (
    "⎕EA", "⎕UCS", "⎕NC", "⎕EX", "⎕SIGNAL", "⎕DR",
    "⎕NREAD", "⎕NWRITE", "⎕NEXISTS", "⎕NDELETE",
    "⎕CR", "⎕FX", "⎕CSV",
)


class Interpreter(Executor):

    def __init__(self, io: int | None = None,
                 fs: FileSystem | None = None,
                 console: 'Console | None' = None) -> None:
        from marple.config import get_default_io
        effective_io = io if io is not None else get_default_io()
        self.env = Environment(io=effective_io, fs=fs, console=console)

    def run(self, source: str) -> APLArray:
        """Parse and evaluate APL source code."""
        if source.strip().startswith("#import"):
            return self._handle_import(source.strip())
        for qfn in _SYS_FUNCTION_NAMES:
            self.env.classify(qfn, NC_FUNCTION)
        source = _newlines_to_diamonds(source)
        tree = parse(source, self.env.class_dict(), self.env.operator_arity_dict())
        result = self.evaluate(tree)
        if isinstance(tree, Assignment):
            self._track_dfn_source(tree.name, source)
        if isinstance(result, DfnBinding):
            return S(0)
        if isinstance(result, APLArray) and is_numeric_array(result.data):
            result = APLArray(list(result.shape), maybe_downcast(result.data, _DOWNCAST_CT))
        return result

    def execute(self, source: str) -> EvalResult:
        """Evaluate APL source and return a structured result.

        Preferred entry point for all interfaces (REPL, web, Jupyter).
        """
        silent = self._is_silent(source)
        value = self.run(source)
        if silent:
            return EvalResult(value=value, silent=True, display_text="")
        text = format_result(value, self.env)
        return EvalResult(value=value, silent=False, display_text=text)

    def _is_silent(self, source: str) -> bool:
        """Check if source is a comment, bare assignment, or directive."""
        stripped = source.strip()
        if stripped.startswith("#") or stripped.startswith("⍝"):
            return True
        try:
            tree = parse(stripped)
        except Exception:
            return False
        if isinstance(tree, Assignment):
            return True
        if isinstance(tree, Program):
            return (len(tree.statements) > 0
                    and isinstance(tree.statements[-1], Assignment))
        return False

    def _handle_import(self, source: str) -> APLArray:
        """Handle #import directive."""
        from marple.errors import DomainError, ValueError_
        from marple.nodes import IBeamDerived
        parts = source.split()
        if len(parts) < 2:
            raise DomainError("Invalid #import directive")
        qualified = parts[1]
        alias = parts[3] if len(parts) >= 4 and parts[2] == "as" else None
        name_parts = qualified.split("::")
        if name_parts[0] != "$":
            raise ValueError_(f"Import from non-system namespace not yet supported: {qualified}")
        result = self.resolve_qualified(name_parts)
        bind_name = alias if alias else name_parts[-1]
        if isinstance(result, (DfnBinding, IBeamDerived)) or (
                hasattr(result, 'dfn') and hasattr(result, 'env')):
            self.env.bind_name(bind_name, result, NC_FUNCTION)
        elif isinstance(result, APLArray):
            from marple.symbol_table import NC_ARRAY
            self.env.bind_name(bind_name, result, NC_ARRAY)
        return S(0)

    def _track_dfn_source(self, name: str, source: str) -> None:
        """Record source text for dfn/dop assignments (for workspace save)."""
        value = self.env.get(name)
        if not isinstance(value, DfnBinding):
            return
        self.env.set_source(name, source.strip())
        if "⍺⍺" not in source and "⍵⍵" not in source:
            return
        self.env.classify(name, NC_OPERATOR)
        self.env.set_operator_arity(name, 2 if "⍵⍵" in source else 1)



"""APL environment (workspace state) for MARPLE."""

from typing import Any

from marple.arraymodel import APLArray, S
from marple.ports.console import Console
from marple.ports.filesystem import FileSystem
from marple.symbol_table import SymbolTable


_QUAD_DEFAULTS: dict[str, APLArray] = {
    "⎕IO": S(1),
    "⎕CT": S(1e-14),
    "⎕PP": S(10),
    "⎕EN": S(0),
    "⎕DM": APLArray([0], []),
    "⎕A": APLArray([26], list("ABCDEFGHIJKLMNOPQRSTUVWXYZ")),
    "⎕D": APLArray([10], list("0123456789")),
    "⎕WSID": APLArray([8], list("CLEAR WS")),
    "⎕RL": S(1),
    "⎕FR": S(645),
    "⎕LX": APLArray([0], []),
}


class Environment:
    """APL workspace environment — variables, system settings, and name table."""

    def __init__(self, io: int | None = None,
                 fs: FileSystem | None = None,
                 console: 'Console | None' = None) -> None:
        self._quad_vars: dict[str, APLArray] = dict(_QUAD_DEFAULTS)
        self.symbols = SymbolTable()
        self._locals: dict[str, Any] = {}
        self.console: Console | None = console
        if io is not None:
            self._quad_vars["⎕IO"] = S(io)
        if fs is not None:
            self.fs = fs
        else:
            from marple.adapters.os_filesystem import OsFileSystem
            self.fs = OsFileSystem()

    # ── System variable properties ──

    @property
    def io(self) -> int:
        return int(self._quad_vars["⎕IO"].data[0])

    @property
    def ct(self) -> float:
        return float(self._quad_vars["⎕CT"].data[0])

    @property
    def pp(self) -> int:
        return int(self._quad_vars["⎕PP"].data[0])

    @property
    def fr(self) -> int:
        return int(self._quad_vars["⎕FR"].data[0])

    # ── Symbol table delegation ──

    def bind_name(self, name: str, value: object, name_class: int) -> None:
        """Store a user-defined name with its value and class."""
        self.symbols.bind(name, value, name_class)

    def classify(self, name: str, name_class: int) -> None:
        """Set a name's class without storing a value (e.g. system functions)."""
        self.symbols.classify(name, name_class)

    def name_class(self, name: str) -> int:
        """Return the name class of a name (0 if unknown)."""
        return self.symbols.name_class(name)

    def delete_name(self, name: str) -> bool:
        """Remove a user-defined name. Returns True if it existed."""
        return self.symbols.delete(name)

    def names_of_class(self, nc: int) -> list[str]:
        """Return sorted list of user names with the given class."""
        return self.symbols.names_of_class(nc)

    def class_dict(self) -> dict[str, int]:
        """Return the raw name-class mapping (for parser compatibility)."""
        return self.symbols.class_dict()

    def set_operator_arity(self, name: str, arity: int) -> None:
        """Record the arity of a user-defined operator."""
        self.symbols.set_operator_arity(name, arity)

    def operator_arity_dict(self) -> dict[str, int]:
        """Return the raw operator arity mapping (for parser compatibility)."""
        return self.symbols.operator_arity_dict()

    def set_source(self, name: str, source: str) -> None:
        """Record the source text of a dfn/dop assignment."""
        self.symbols.set_source(name, source)

    def get_source(self, name: str) -> str | None:
        """Return the source text for a name, or None."""
        return self.symbols.get_source(name)

    def sources(self) -> dict[str, str]:
        """Return the raw sources mapping."""
        return self.symbols.sources()

    def quad_var_names(self) -> list[str]:
        """Return sorted list of quad variable names."""
        return sorted(self._quad_vars.keys())

    def user_names(self) -> list[str]:
        """Return sorted list of user-defined names in the symbol table."""
        return sorted(n for n in self.symbols._values.keys()
                       if not n.startswith("__"))

    def list_variables(self) -> list[tuple[str, list[int]]]:
        """Return sorted list of (name, shape) for all user variables."""
        result = []
        for name in self.names_of_class(2):  # NC_ARRAY
            val = self.symbols.get(name)
            if isinstance(val, APLArray):
                result.append((name, list(val.shape)))
        return result

    def list_functions(self) -> list[tuple[str, str | None]]:
        """Return sorted list of (name, source_or_None) for all user functions."""
        result = []
        for name in self.names_of_class(3):  # NC_FUNCTION
            source = self.get_source(name)
            result.append((name, source))
        return result

    def list_operators(self) -> list[tuple[str, str | None]]:
        """Return sorted list of (name, source_or_None) for all user operators."""
        result = []
        for name in self.names_of_class(4):  # NC_OPERATOR
            source = self.get_source(name)
            result.append((name, source))
        return result

    # ── Dict-like interface ──
    # Lookup order: quad vars, then symbols, then locals (⍵, ⍺, ∇, etc.)

    def __getitem__(self, key: str) -> Any:
        if key in self._quad_vars:
            return self._quad_vars[key]
        val = self.symbols.get(key)
        if val is not None:
            return val
        return self._locals[key]

    def __setitem__(self, key: str, value: Any) -> None:
        if key in self._quad_vars:
            self._quad_vars[key] = value
        else:
            self._locals[key] = value

    def __contains__(self, key: object) -> bool:
        if not isinstance(key, str):
            return False
        return key in self._quad_vars or key in self.symbols or key in self._locals

    def __delitem__(self, key: str) -> None:
        if key in self._locals:
            del self._locals[key]

    def get(self, key: str, default: Any = None) -> Any:
        if key in self._quad_vars:
            return self._quad_vars[key]
        val = self.symbols.get(key)
        if val is not None:
            return val
        return self._locals.get(key, default)

    def setdefault(self, key: str, default: Any) -> Any:
        return self._locals.setdefault(key, default)

    def pop(self, key: str, *args: Any) -> Any:
        return self._locals.pop(key, *args)

    def items(self) -> Any:
        return self._locals.items()

    def copy(self) -> 'Environment':
        """Copy for dfn local environments.

        Quad vars, symbols, and locals are all copied so that
        assignments inside a dfn do not leak to the caller.
        """
        new = Environment(fs=self.fs, console=self.console)
        new._quad_vars = dict(self._quad_vars)
        new.symbols = self.symbols.copy()
        new._locals = dict(self._locals)
        return new



"""DfnBinding: a dfn/dop that can evaluate its own body."""

from __future__ import annotations

from typing import TYPE_CHECKING

from marple.arraymodel import APLArray, S
from marple.executor import Executor
from marple.parser import AlphaDefault, Dfn, Guard

if TYPE_CHECKING:
    from marple.environment import Environment


class _TailCall:
    """Signal: re-enter dfn with new args instead of recursing."""
    __slots__ = ("omega", "alpha")

    def __init__(self, omega: APLArray, alpha: APLArray | None) -> None:
        self.omega = omega
        self.alpha = alpha


def _is_tail_self_call(node: object) -> bool:
    """Check if an AST node is a ∇ call in tail position."""
    from marple.nodes import DyadicDfnCall, MonadicDfnCall, Nabla
    if isinstance(node, MonadicDfnCall) and isinstance(node.dfn, Nabla):
        return True
    if isinstance(node, DyadicDfnCall) and isinstance(node.dfn, Nabla):
        return True
    return False


class DfnBinding(Executor):
    """A dfn or dop bound to the environment in which it was defined."""

    def __init__(self, dfn: Dfn, defining_env: Environment) -> None:
        self.dfn = dfn
        self.defining_env = defining_env
        self.env = defining_env  # current env; swapped during apply

    def apply(
        self,
        omega: APLArray,
        alpha: APLArray | None = None,
        alpha_alpha: object | None = None,
        omega_omega: object | None = None,
    ) -> APLArray:
        """Apply this dfn or dop to its arguments."""
        saved_env = self.env
        try:
            while True:
                self.env = self._make_env(omega, alpha, alpha_alpha, omega_omega)
                result = self._execute_body()
                if isinstance(result, _TailCall):
                    omega = result.omega
                    alpha = result.alpha
                    continue
                return result
        finally:
            self.env = saved_env

    def _make_env(
        self,
        omega: APLArray,
        alpha: APLArray | None,
        alpha_alpha: object | None,
        omega_omega: object | None,
    ) -> Environment:
        """Build the local environment for this call."""
        local_env = self.defining_env.copy()
        local_env["⍵"] = omega
        if alpha is not None:
            local_env["⍺"] = alpha
        if alpha_alpha is not None:
            local_env["⍺⍺"] = alpha_alpha
        if omega_omega is not None:
            local_env["⍵⍵"] = omega_omega
        local_env["∇"] = self
        return local_env

    def _execute_body(self) -> APLArray | _TailCall:
        """Execute the dfn body statements.

        Returns _TailCall to signal TCO, otherwise the result APLArray.
        """
        from marple.nodes import DyadicDfnCall, MonadicDfnCall
        result: APLArray = S(0)
        stmts = self.dfn.body
        last_idx = len(stmts) - 1
        for idx, stmt in enumerate(stmts):
            if isinstance(stmt, AlphaDefault):
                if "⍺" not in self.env:
                    self.env["⍺"] = self.evaluate(stmt.default)
            elif isinstance(stmt, Guard):
                cond = self.evaluate(stmt.condition)
                if cond.data[0]:
                    if _is_tail_self_call(stmt.body):
                        return self._make_tail_call(stmt.body)
                    return self.evaluate(stmt.body)
            else:
                if idx == last_idx and _is_tail_self_call(stmt):
                    return self._make_tail_call(stmt)
                result = self.evaluate(stmt)
        return result

    def _make_tail_call(self, node: object) -> _TailCall:
        """Extract args from a tail-position ∇ call and return a _TailCall signal."""
        from marple.nodes import DyadicDfnCall, MonadicDfnCall
        if isinstance(node, MonadicDfnCall):
            omega = self.evaluate(node.operand)
            return _TailCall(omega, None)
        if isinstance(node, DyadicDfnCall):
            omega = self.evaluate(node.right)
            alpha = self.evaluate(node.left)
            return _TailCall(omega, alpha)
        raise RuntimeError("_make_tail_call called on non-∇ node")



"""FileSystem port — abstract interface for file I/O."""

from abc import ABC, abstractmethod


class FileSystem(ABC):
    """Port for filesystem operations."""

    @abstractmethod
    def read_text(self, path: str) -> str:
        """Read a file and return its contents as a string."""
        ...

    @abstractmethod
    def write_text(self, path: str, content: str) -> None:
        """Write a string to a file, creating or overwriting it."""
        ...

    @abstractmethod
    def exists(self, path: str) -> bool:
        """Return True if the path exists (file or directory)."""
        ...

    @abstractmethod
    def is_file(self, path: str) -> bool:
        """Return True if the path is an existing file."""
        ...

    @abstractmethod
    def is_dir(self, path: str) -> bool:
        """Return True if the path is an existing directory."""
        ...

    @abstractmethod
    def delete(self, path: str) -> None:
        """Delete a file. Raises OSError if not found."""
        ...

    @abstractmethod
    def makedirs(self, path: str) -> None:
        """Create a directory and any missing parents."""
        ...

    @abstractmethod
    def listdir(self, path: str) -> list[str]:
        """List entries in a directory."""
        ...



"""Symbol table for user-defined names in MARPLE."""

from typing import Any

# Name classes (following Dyalog ⎕NC convention)
NC_UNKNOWN = 0
NC_ARRAY = 2
NC_FUNCTION = 3
NC_OPERATOR = 4


class SymbolTable:
    """Tracks user-defined names, their values, and their name classes."""

    def __init__(self) -> None:
        self._values: dict[str, Any] = {}
        self._classes: dict[str, int] = {}
        self._operator_arity: dict[str, int] = {}
        self._sources: dict[str, str] = {}

    def bind(self, name: str, value: Any, name_class: int) -> None:
        """Store a name with its value and class."""
        self._values[name] = value
        self._classes[name] = name_class

    def classify(self, name: str, name_class: int) -> None:
        """Set a name's class without storing a value (e.g. system functions)."""
        self._classes[name] = name_class

    def get(self, name: str) -> Any:
        """Return the value bound to a name, or None if not found."""
        return self._values.get(name)

    def name_class(self, name: str) -> int:
        """Return the name class (0 if unknown)."""
        return self._classes.get(name, NC_UNKNOWN)

    def __contains__(self, name: object) -> bool:
        return name in self._values

    def set_operator_arity(self, name: str, arity: int) -> None:
        """Record the arity (1=monadic, 2=dyadic) of a user-defined operator."""
        self._operator_arity[name] = arity

    def operator_arity_dict(self) -> dict[str, int]:
        """Return the raw operator arity mapping (for parser compatibility)."""
        return self._operator_arity

    def set_source(self, name: str, source: str) -> None:
        """Record the source text of a dfn/dop assignment."""
        self._sources[name] = source

    def get_source(self, name: str) -> str | None:
        """Return the source text for a name, or None."""
        return self._sources.get(name)

    def sources(self) -> dict[str, str]:
        """Return the raw sources mapping."""
        return self._sources

    def delete(self, name: str) -> bool:
        """Remove a name. Returns True if it existed."""
        found = name in self._values
        self._values.pop(name, None)
        self._classes.pop(name, None)
        self._operator_arity.pop(name, None)
        self._sources.pop(name, None)
        return found

    def names_of_class(self, nc: int) -> list[str]:
        """Return sorted list of user names with the given class."""
        return sorted(
            n for n, c in self._classes.items()
            if c == nc and not n.startswith("⎕") and not n.startswith("__")
        )

    def class_dict(self) -> dict[str, int]:
        """Return the raw name-class mapping (for parser compatibility)."""
        return self._classes

    def copy(self) -> 'SymbolTable':
        """Shallow copy of this symbol table."""
        new = SymbolTable()
        new._values = dict(self._values)
        new._classes = dict(self._classes)
        new._operator_arity = dict(self._operator_arity)
        new._sources = dict(self._sources)
        return new



"""Install the MARPLE Jupyter kernelspec."""

import json
import shutil
import tempfile
from pathlib import Path

from jupyter_client.kernelspec import KernelSpecManager

KERNEL_JSON = {
    "argv": ["python", "-m", "marple.jupyter", "-f", "{connection_file}"],
    "display_name": "MARPLE (APL)",
    "language": "apl",
    "interrupt_mode": "signal",
}


def install_kernel(user: bool = True) -> None:
    """Install the MARPLE kernelspec."""
    resources = Path(__file__).parent / "resources"
    with tempfile.TemporaryDirectory() as td:
        td_path = Path(td)
        (td_path / "kernel.json").write_text(json.dumps(KERNEL_JSON, indent=2))
        for fname in ["logo-32x32.png", "logo-64x64.png", "kernel.js"]:
            src = resources / fname
            if src.exists():
                shutil.copy(src, td_path)
        KernelSpecManager().install_kernel_spec(
            str(td_path), "marple", user=user)
    print("MARPLE kernel installed. Run: jupyter notebook")


if __name__ == "__main__":
    install_kernel()



from typing import Callable

from marple.errors import SyntaxError_
from marple.nodes import (  # noqa: F401 — re-exported for backward compatibility
    Alpha,
    AlphaAlpha,
    AlphaDefault,
    Assignment,
    BoundOperator,
    CAT_EMPTY,
    DerivedFunc,
    Dfn,
    DyadicDfnCall,
    DyadicDopCall,
    DyadicFunc,
    ExecutionContext,
    FmtArgs,
    FunctionRef,
    Guard,
    IBeamDerived,
    Index,
    InnerProduct,
    MonadicDfnCall,
    MonadicDopCall,
    MonadicFunc,
    Nabla,
    Node,
    Num,
    Omega,
    OmegaOmega,
    OuterProduct,
    PowerDerived,
    Program,
    QualifiedVar,
    RankDerived,
    ReduceOp,
    ScanOp,
    Str,
    SysVar,
    Var,
    Vector,
)
from marple.tokenizer import Token, TokenType, Tokenizer


# ── Category constants for Iverson's stack-based parser ──

CAT_END = 0    # End marker (⋄, EOF)
CAT_NOUN = 1   # Noun (array/data)
CAT_VERB = 2   # Verb (function)
CAT_ADV = 3    # Adverb (monadic operator: /, \, ⌿, ⍀, ∘.)
CAT_CONJ = 4   # Conjunction (dyadic operator: ⍤, ., ⌶)
CAT_ASGN = 5   # Assignment (←)
CAT_LP = 6     # Left paren
CAT_RP = 7     # Right paren
# CAT_EMPTY imported from nodes

# Context sets for case matching
_CTX_MONAD = frozenset({CAT_END, CAT_ADV, CAT_VERB, CAT_ASGN, CAT_LP})
_CTX_DYAD = frozenset({CAT_END, CAT_NOUN, CAT_ADV, CAT_VERB, CAT_ASGN, CAT_LP})

# System functions (classified as verbs) vs system variables (nouns)
_SYS_FUNCTIONS = frozenset({
    "⎕CR", "⎕FX", "⎕NC", "⎕EX", "⎕SIGNAL", "⎕EA",
    "⎕UCS", "⎕DR", "⎕NREAD", "⎕NWRITE", "⎕NEXISTS", "⎕NDELETE",
    "⎕DL", "⎕FMT", "⎕VFI", "⎕JSON", "⎕NL", "⎕CSV",
})


def _ast_contains(node: object, target_type: type) -> bool:
    """Check if any node in an AST subtree is of target_type."""
    if isinstance(node, target_type):
        return True
    if hasattr(node, '__dict__'):
        for val in node.__dict__.values():
            if isinstance(val, list):
                if any(_ast_contains(item, target_type) for item in val):
                    return True
            elif hasattr(val, '__dict__') and not isinstance(val, type):
                if _ast_contains(val, target_type):
                    return True
    return False


class Parser:
    """Stack-based APL parser following Iverson's parsing algorithm.

    Uses a 4-position window on a stack with 9 pattern-matching rules
    to correctly handle operator binding precedence.
    """

    def __init__(self, tokens: list[Token], name_table: dict[str, int] | None = None,
                 operator_arity: dict[str, int] | None = None) -> None:
        self._tokens = tokens
        self._pos = 0
        self._name_table = name_table or {}
        self._operator_arity = operator_arity or {}

    def _is_function_name(self, name: str) -> bool:
        """Check if a name is classified as a function in the name table."""
        return self._name_table.get(name) == 3  # NC_FUNCTION

    def _is_operator_name(self, name: str) -> bool:
        """Check if a name is classified as an operator in the name table."""
        return self._name_table.get(name) == 4  # NC_OPERATOR

    # ── Token classification ──

    def _classify_operator(self, op: str) -> int:
        """Classify an operator token as adverb or conjunction."""
        if op in ("/", "\\", "⌿", "⍀", "∘."):
            return CAT_ADV
        if op in ("⍤", "⍣", "⌶", ".", "∘"):
            return CAT_CONJ
        return CAT_ADV

    def _classify_name(self, name: str) -> int:
        """Classify a name using the name table."""
        if self._is_operator_name(name):
            arity = self._operator_arity.get(name, 1)
            return CAT_ADV if arity == 1 else CAT_CONJ
        if self._is_function_name(name):
            return CAT_VERB
        return CAT_NOUN

    def _classify_sysvar(self, name: str) -> int:
        """Classify a system variable/function."""
        if name in _SYS_FUNCTIONS:
            return CAT_VERB
        return CAT_NOUN

    def _classify_dfn(self, dfn: Dfn) -> int:
        """Classify a dfn as verb, adverb, or conjunction."""
        if _ast_contains(dfn, OmegaOmega):
            return CAT_CONJ
        if _ast_contains(dfn, AlphaAlpha):
            return CAT_ADV
        return CAT_VERB

    # ── Item building (token → classified items) ──

    def _build_items(self) -> list[tuple[int, object]]:
        """Build classified items from current position to expression end.

        Advances self._pos as tokens are consumed. Stops at statement
        boundaries (⋄, :, }, ], ;, EOF) at paren depth 0.
        """
        items: list[tuple[int, object]] = []
        paren_depth = 0
        stop_types = frozenset({
            TokenType.DIAMOND, TokenType.GUARD, TokenType.RBRACE,
            TokenType.RBRACKET, TokenType.SEMICOLON, TokenType.EOF,
            TokenType.LBRACKET,
        })

        while self._pos < len(self._tokens):
            tok = self._current()

            if paren_depth == 0 and tok.type in stop_types:
                break

            # Parens need inline handling — they track paren_depth state
            if tok.type == TokenType.LPAREN:
                items.append((CAT_LP, None))
                self._pos += 1
                paren_depth += 1
                continue
            if tok.type == TokenType.RPAREN:
                paren_depth -= 1
                if paren_depth < 0:
                    break
                items.append((CAT_RP, None))
                self._pos += 1
                # (expr)[idx] — bracket index binds to parenthesised expression
                if (paren_depth == 0
                        and self._pos < len(self._tokens)
                        and self._current().type == TokenType.LBRACKET):
                    # Reduce the paren group we just closed, then apply [idx]
                    lp_pos = len(items) - 1
                    while lp_pos >= 0 and items[lp_pos][0] != CAT_LP:
                        lp_pos -= 1
                    if lp_pos >= 0:
                        paren_items = items[lp_pos + 1:-1]  # between LP and RP
                        inner = self._stack_parse(paren_items)
                        node = self._parse_bracket_index(inner)
                        items[lp_pos:] = [(CAT_NOUN, node)]
                continue

            handler = self._ITEM_DISPATCH.get(tok.type)
            if handler is not None:
                handler(self, tok, items)
            else:
                raise SyntaxError_(f"Unexpected token: {tok}")

        return items

    def _item_lbrace(self, tok: Token, items: list[tuple[int, object]]) -> None:
        dfn = self._parse_dfn()
        cat = self._classify_dfn(dfn)
        items.append((cat, dfn))

    def _item_number(self, tok: Token, items: list[tuple[int, object]]) -> None:
        node: object = self._parse_array()
        if self._current().type == TokenType.LBRACKET:
            node = self._parse_bracket_index(node)
        items.append((CAT_NOUN, node))

    def _item_string(self, tok: Token, items: list[tuple[int, object]]) -> None:
        self._pos += 1
        assert isinstance(tok.value, str)
        node: object = Str(tok.value)
        if self._current().type == TokenType.LBRACKET:
            node = self._parse_bracket_index(node)
        items.append((CAT_NOUN, node))

    def _item_function(self, tok: Token, items: list[tuple[int, object]]) -> None:
        self._pos += 1
        assert isinstance(tok.value, str)
        items.append((CAT_VERB, tok.value))

    def _item_operator(self, tok: Token, items: list[tuple[int, object]]) -> None:
        assert isinstance(tok.value, str)
        op = tok.value
        self._pos += 1
        if (op == "∘" and self._pos < len(self._tokens)
                and self._current().type == TokenType.OPERATOR
                and self._current().value == "."):
            self._pos += 1
            fn_tok = self._eat(TokenType.FUNCTION)
            assert isinstance(fn_tok.value, str)
            bound = BoundOperator("∘.", fn_tok.value, CAT_VERB)
            items.append((CAT_VERB, bound))
        elif (op == "⌶" and self._pos < len(self._tokens)
                and self._current().type == TokenType.STRING):
            path_tok = self._eat(TokenType.STRING)
            assert isinstance(path_tok.value, str)
            items.append((CAT_VERB, IBeamDerived(path_tok.value)))
        else:
            cat = self._classify_operator(op)
            items.append((cat, op))

    def _item_assign(self, tok: Token, items: list[tuple[int, object]]) -> None:
        self._pos += 1
        items.append((CAT_ASGN, "←"))

    def _item_id(self, tok: Token, items: list[tuple[int, object]]) -> None:
        assert isinstance(tok.value, str)
        name = tok.value
        self._pos += 1
        var_node = Var(name)
        if self._current().type == TokenType.LBRACKET:
            node = self._parse_bracket_index(var_node)
            items.append((CAT_NOUN, node))
        elif self._current().type == TokenType.ASSIGN:
            items.append((CAT_NOUN, var_node))
        else:
            cat = self._classify_name(name)
            items.append((cat, var_node))

    def _item_sysvar(self, tok: Token, items: list[tuple[int, object]]) -> None:
        assert isinstance(tok.value, str)
        name = tok.value
        self._pos += 1
        sv_node = SysVar(name)
        if self._current().type == TokenType.LBRACKET:
            node = self._parse_bracket_index(sv_node)
            items.append((CAT_NOUN, node))
        elif self._current().type == TokenType.ASSIGN:
            items.append((CAT_NOUN, sv_node))
        elif (name == "⎕FMT"
              and self._current().type == TokenType.LPAREN):
            items.append((CAT_VERB, sv_node))
            fmt_args = self._parse_fmt_args()
            items.append((CAT_NOUN, fmt_args))
        else:
            cat = self._classify_sysvar(name)
            items.append((cat, sv_node))

    def _item_omega(self, tok: Token, items: list[tuple[int, object]]) -> None:
        self._pos += 1
        items.append((CAT_NOUN, Omega()))

    def _item_alpha(self, tok: Token, items: list[tuple[int, object]]) -> None:
        self._pos += 1
        items.append((CAT_NOUN, Alpha()))

    def _item_alpha_alpha(self, tok: Token, items: list[tuple[int, object]]) -> None:
        self._pos += 1
        items.append((CAT_NOUN, AlphaAlpha()))

    def _item_omega_omega(self, tok: Token, items: list[tuple[int, object]]) -> None:
        self._pos += 1
        items.append((CAT_NOUN, OmegaOmega()))

    def _item_nabla(self, tok: Token, items: list[tuple[int, object]]) -> None:
        self._pos += 1
        items.append((CAT_VERB, Nabla()))

    def _item_qualified_name(self, tok: Token, items: list[tuple[int, object]]) -> None:
        assert isinstance(tok.value, str)
        self._pos += 1
        items.append((CAT_NOUN, QualifiedVar(tok.value.split("::"))))

    _ITEM_DISPATCH: dict[str, Callable[['Parser', Token, list[tuple[int, object]]], None]] = {
        TokenType.LBRACE: _item_lbrace,
        TokenType.NUMBER: _item_number,
        TokenType.STRING: _item_string,
        TokenType.FUNCTION: _item_function,
        TokenType.OPERATOR: _item_operator,
        TokenType.ASSIGN: _item_assign,
        TokenType.ID: _item_id,
        TokenType.SYSVAR: _item_sysvar,
        TokenType.OMEGA: _item_omega,
        TokenType.ALPHA: _item_alpha,
        TokenType.ALPHA_ALPHA: _item_alpha_alpha,
        TokenType.OMEGA_OMEGA: _item_omega_omega,
        TokenType.NABLA: _item_nabla,
        TokenType.QUALIFIED_NAME: _item_qualified_name,
    }

    def _parse_fmt_args(self) -> FmtArgs:
        """Parse (val1;val2;...) for ⎕FMT. Similar to bracket indexing."""
        self._eat(TokenType.LPAREN)
        args: list[object] = []
        args.append(self._parse_statement())
        while self._current().type == TokenType.SEMICOLON:
            self._eat(TokenType.SEMICOLON)
            args.append(self._parse_statement())
        self._eat(TokenType.RPAREN)
        return FmtArgs(args)

    def _is_callable_noun(self, node: object) -> bool:
        """Check if a NOUN-classified node can act as a function.
        Returns True for: ⍺⍺, ⍵⍵, ∇, all Vars (may be dynamically-defined
        functions), QualifiedVars, system functions, Dfns."""
        if isinstance(node, (AlphaAlpha, OmegaOmega, Nabla, Dfn)):
            return True
        if isinstance(node, (Var, QualifiedVar)):
            return True
        if isinstance(node, SysVar) and node.name in _SYS_FUNCTIONS:
            return True
        return False

    # ── AST construction helpers ──

    def _make_monadic(self, verb_node: object, arg_node: object) -> object:
        """Create AST node for monadic verb application."""
        if isinstance(verb_node, str):
            return MonadicFunc(verb_node, arg_node)
        if isinstance(verb_node, BoundOperator):
            return self._apply_bound_monadic(verb_node, arg_node)
        if isinstance(verb_node, (Var, Dfn, QualifiedVar, Nabla,
                                  AlphaAlpha, OmegaOmega,
                                  RankDerived, IBeamDerived, FunctionRef)):
            return MonadicDfnCall(verb_node, arg_node)
        if isinstance(verb_node, SysVar):
            return MonadicDfnCall(verb_node, arg_node)
        raise SyntaxError_(f"Cannot apply as monadic function: {type(verb_node)}")

    def _make_dyadic(self, verb_node: object, left_node: object,
                     right_node: object) -> object:
        """Create AST node for dyadic verb application."""
        if isinstance(verb_node, str):
            return DyadicFunc(verb_node, left_node, right_node)
        if isinstance(verb_node, BoundOperator):
            return self._apply_bound_dyadic(verb_node, left_node, right_node)
        if isinstance(verb_node, (Var, Dfn, QualifiedVar, Nabla,
                                  AlphaAlpha, OmegaOmega,
                                  RankDerived, IBeamDerived, FunctionRef)):
            return DyadicDfnCall(verb_node, left_node, right_node)
        if isinstance(verb_node, SysVar):
            return DyadicDfnCall(verb_node, left_node, right_node)
        raise SyntaxError_(f"Cannot apply as dyadic function: {type(verb_node)}")

    def _apply_bound_monadic(self, bound: BoundOperator,
                             arg_node: object) -> object:
        """Apply a bound operator (derived verb) monadically."""
        op = bound.operator
        if isinstance(op, Var):
            return self._apply_user_dop_monadic(bound, arg_node)
        if not isinstance(op, str):
            raise SyntaxError_(f"Unknown operator in bound form: {op}")
        handler = self._BOUND_MONADIC_DISPATCH.get(op)
        if handler is not None:
            return handler(self, bound, arg_node)
        raise SyntaxError_(f"Unknown operator in bound form: {op}")

    def _bound_monadic_reduce(self, bound: BoundOperator, arg_node: object) -> object:
        assert isinstance(bound.operator, str)
        operand = bound.left_operand
        if (bound.left_cat == CAT_VERB
                or isinstance(operand, (AlphaAlpha, OmegaOmega))):
            return DerivedFunc(bound.operator, operand, arg_node)
        return DyadicFunc(bound.operator, operand, arg_node)

    def _bound_monadic_rank(self, bound: BoundOperator, arg_node: object) -> object:
        rank_node = RankDerived(bound.left_operand, bound.right_operand)
        return MonadicDfnCall(rank_node, arg_node)

    def _bound_monadic_power(self, bound: BoundOperator, arg_node: object) -> object:
        power_node = PowerDerived(bound.left_operand, bound.right_operand)
        return MonadicDfnCall(power_node, arg_node)

    def _bound_monadic_inner(self, bound: BoundOperator, arg_node: object) -> object:
        raise SyntaxError_("Inner product requires two arguments")

    def _bound_monadic_outer(self, bound: BoundOperator, arg_node: object) -> object:
        raise SyntaxError_("Outer product requires two arguments")

    def _bound_monadic_ibeam(self, bound: BoundOperator, arg_node: object) -> object:
        operand = bound.left_operand
        ibeam = IBeamDerived(operand) if isinstance(operand, str) else operand
        return MonadicDfnCall(ibeam, arg_node)

    def _apply_user_dop_monadic(self, bound: BoundOperator, arg_node: object) -> object:
        operand = bound.left_operand
        op_operand = FunctionRef(operand) if isinstance(operand, str) else operand
        if bound.right_operand is not None:
            r_operand = bound.right_operand
            r_operand = FunctionRef(r_operand) if isinstance(r_operand, str) else r_operand
            return DyadicDopCall(bound.operator, op_operand, r_operand, arg_node)
        return MonadicDopCall(bound.operator, op_operand, arg_node)

    _BOUND_MONADIC_DISPATCH: dict[str, Callable[['Parser', BoundOperator, object], object]] = {
        "/": _bound_monadic_reduce,
        "\\": _bound_monadic_reduce,
        "⌿": _bound_monadic_reduce,
        "⍀": _bound_monadic_reduce,
        "⍤": _bound_monadic_rank,
        "⍣": _bound_monadic_power,
        ".": _bound_monadic_inner,
        "∘.": _bound_monadic_outer,
        "⌶": _bound_monadic_ibeam,
    }

    def _apply_bound_dyadic(self, bound: BoundOperator,
                            left_node: object, right_node: object) -> object:
        """Apply a bound operator (derived verb) dyadically."""
        op = bound.operator
        if isinstance(op, Var):
            return self._apply_user_dop_dyadic(bound, left_node, right_node)
        if not isinstance(op, str):
            raise SyntaxError_(f"Unknown operator in bound dyadic form: {op}")
        handler = self._BOUND_DYADIC_DISPATCH.get(op)
        if handler is not None:
            return handler(self, bound, left_node, right_node)
        raise SyntaxError_(f"Unknown operator in bound dyadic form: {op}")

    def _bound_dyadic_rank(self, bound: BoundOperator,
                           left_node: object, right_node: object) -> object:
        rank_node = RankDerived(bound.left_operand, bound.right_operand)
        return DyadicDfnCall(rank_node, left_node, right_node)

    def _bound_dyadic_power(self, bound: BoundOperator,
                            left_node: object, right_node: object) -> object:
        power_node = PowerDerived(bound.left_operand, bound.right_operand)
        return DyadicDfnCall(power_node, left_node, right_node)

    def _bound_dyadic_inner(self, bound: BoundOperator,
                            left_node: object, right_node: object) -> object:
        assert bound.right_operand is not None
        return InnerProduct(bound.left_operand, bound.right_operand,
                            left_node, right_node)

    def _bound_dyadic_outer(self, bound: BoundOperator,
                            left_node: object, right_node: object) -> object:
        return OuterProduct(bound.left_operand, left_node, right_node)

    def _bound_dyadic_reduce(self, bound: BoundOperator,
                             left_node: object, right_node: object) -> object:
        assert isinstance(bound.operator, str)
        operand = bound.left_operand
        if (bound.left_cat == CAT_VERB
                or isinstance(operand, (AlphaAlpha, OmegaOmega))):
            return DerivedFunc(bound.operator, operand, right_node)
        return DyadicFunc(bound.operator, operand, right_node)

    def _apply_user_dop_dyadic(self, bound: BoundOperator,
                               left_node: object, right_node: object) -> object:
        operand = bound.left_operand
        op_operand = FunctionRef(operand) if isinstance(operand, str) else operand
        if bound.right_operand is not None:
            r_operand = bound.right_operand
            r_operand = FunctionRef(r_operand) if isinstance(r_operand, str) else r_operand
            return DyadicDopCall(bound.operator, op_operand, r_operand, right_node)
        return MonadicDopCall(bound.operator, op_operand, right_node, alpha=left_node)

    _BOUND_DYADIC_DISPATCH: dict[str, Callable[['Parser', BoundOperator, object, object], object]] = {
        "⍤": _bound_dyadic_rank,
        "⍣": _bound_dyadic_power,
        ".": _bound_dyadic_inner,
        "∘.": _bound_dyadic_outer,
        "/": _bound_dyadic_reduce,
        "\\": _bound_dyadic_reduce,
        "⌿": _bound_dyadic_reduce,
        "⍀": _bound_dyadic_reduce,
    }

    # ── Iverson's stack-based parsing algorithm ──

    def _stack_parse(self, items: list[tuple[int, object]]) -> object:
        """Run Iverson's 9-case stack-based parser on classified items.

        Items are processed right-to-left. The stack grows upward.
        stack[-1] = r0 (top/newest), stack[-2] = r1, etc.
        """
        stack: list[tuple[int, object]] = []
        # Items in left-to-right order; pop from right = read right-to-left
        # END marker at position 0 = popped last = leftmost sentinel
        input_q = [(CAT_END, None)] + items

        while True:
            n = len(stack)
            c0 = stack[-1][0] if n >= 1 else CAT_EMPTY
            c1 = stack[-2][0] if n >= 2 else CAT_EMPTY
            c2 = stack[-3][0] if n >= 3 else CAT_EMPTY
            c3 = stack[-4][0] if n >= 4 else CAT_EMPTY

            matched = False

            # Case 1: Monadic function — E/A/V/←/LP  V  N  —
            if (c0 in _CTX_MONAD
                    and c1 == CAT_VERB and c2 == CAT_NOUN):
                verb_node = stack[-2][1]
                arg_node = stack[-3][1]
                result = self._make_monadic(verb_node, arg_node)
                stack[-3:-1] = [(CAT_NOUN, result)]
                matched = True

            # Case 1.5: Callable noun as monadic function
            # Handles ⍺⍺, ⍵⍵, ∇, named functions, Vars applied monadically
            # Skip when deeper callable+noun exists (let case 3.6 handle it)
            elif (c0 in _CTX_MONAD
                    and c1 == CAT_NOUN and c2 == CAT_NOUN
                    and self._is_callable_noun(stack[-2][1])
                    and not (n >= 4 and c3 == CAT_NOUN
                             and self._is_callable_noun(stack[-3][1]))):
                verb_node = stack[-2][1]
                arg_node = stack[-3][1]
                result = self._make_monadic(verb_node, arg_node)
                stack[-3:-1] = [(CAT_NOUN, result)]
                matched = True

            # Case 2: Monadic fn (conjunction context) — C  N  V  N
            elif (c0 == CAT_CONJ and c1 == CAT_NOUN
                    and c2 == CAT_VERB and c3 == CAT_NOUN):
                verb_node = stack[-3][1]
                arg_node = stack[-4][1]
                result = self._make_monadic(verb_node, arg_node)
                stack[-4:-2] = [(CAT_NOUN, result)]
                matched = True

            # Case 3: Dyadic function — E/N/A/V/←/LP  N  V  N
            elif (c0 in _CTX_DYAD
                    and c1 == CAT_NOUN and c2 == CAT_VERB
                    and c3 == CAT_NOUN):
                left_node = stack[-2][1]
                verb_node = stack[-3][1]
                right_node = stack[-4][1]
                result = self._make_dyadic(verb_node, left_node, right_node)
                stack[-4:-1] = [(CAT_NOUN, result)]
                matched = True

            # Case 3.5: Dyadic callable noun — N(non-callable) N(callable) N
            # E.g. `3 add 4` where add is a dynamically-defined function
            elif (c0 in _CTX_DYAD
                    and c1 == CAT_NOUN and c2 == CAT_NOUN
                    and c3 == CAT_NOUN
                    and not self._is_callable_noun(stack[-2][1])
                    and self._is_callable_noun(stack[-3][1])):
                left_node = stack[-2][1]
                verb_node = stack[-3][1]
                right_node = stack[-4][1]
                result = self._make_dyadic(verb_node, left_node, right_node)
                stack[-4:-1] = [(CAT_NOUN, result)]
                matched = True

            # Case 3.6: Deep monadic chain — N(callable) N(callable) N
            # E.g. `⍺⍺ ⍺⍺ ⍵`: bind deeper pair first for right-to-left eval
            elif (c0 in _CTX_DYAD
                    and c2 == CAT_NOUN and c3 == CAT_NOUN
                    and self._is_callable_noun(stack[-3][1])):
                verb_node = stack[-3][1]
                arg_node = stack[-4][1]
                result = self._make_monadic(verb_node, arg_node)
                stack[-4:-2] = [(CAT_NOUN, result)]
                matched = True

            # Case 4: Adverb binding — E/N/A/V/←/LP  N/V  A  —
            elif (c0 in _CTX_DYAD
                    and c1 in (CAT_NOUN, CAT_VERB) and c2 == CAT_ADV):
                operand_node = stack[-2][1]
                operand_cat = stack[-2][0]
                adv_node = stack[-3][1]
                assert isinstance(adv_node, (str, Var))
                assert isinstance(operand_node, (str, Node, BoundOperator))
                bound = BoundOperator(adv_node, operand_node, operand_cat)
                stack[-3:-1] = [(CAT_VERB, bound)]
                matched = True

            # Case 5: Conjunction binding — E/N/A/V/←/LP  N/V  C  N/V
            elif (c0 in _CTX_DYAD
                    and c1 in (CAT_NOUN, CAT_VERB) and c2 == CAT_CONJ
                    and c3 in (CAT_NOUN, CAT_VERB)):
                left_operand = stack[-2][1]
                left_cat = stack[-2][0]
                conj_node = stack[-3][1]
                right_operand = stack[-4][1]
                right_cat = stack[-4][0]
                assert isinstance(conj_node, (str, Var))
                assert isinstance(left_operand, (str, Node, BoundOperator))
                assert isinstance(right_operand, (str, Node, BoundOperator))
                bound = BoundOperator(conj_node, left_operand, left_cat,
                                      right_operand, right_cat)
                stack[-4:-1] = [(CAT_VERB, bound)]
                matched = True

            # Case 6: Assignment — N/name  ←  N/V/A/C  —
            elif (c0 in (CAT_NOUN,) and c1 == CAT_ASGN
                    and c2 in (CAT_NOUN, CAT_VERB, CAT_ADV, CAT_CONJ)):
                name_node = stack[-1][1]
                value_node = stack[-3][1]
                value_cat = stack[-3][0]
                if isinstance(name_node, (Var, SysVar)):
                    name_str = name_node.name
                else:
                    raise SyntaxError_(f"Invalid assignment target: {name_node}")
                result = Assignment(name_str, value_node)
                stack[-3:] = [(value_cat, result)]
                matched = True

            # Case 7: Parentheses — LP  N/A/C/V  RP  —
            elif (c0 == CAT_LP
                    and c1 in (CAT_NOUN, CAT_ADV, CAT_CONJ, CAT_VERB)
                    and c2 == CAT_RP):
                inner_cat = stack[-2][0]
                inner_node = stack[-2][1]
                stack[-3:] = [(inner_cat, inner_node)]
                matched = True

            if not matched:
                # Cases 8/9: Shift next item from input
                if input_q:
                    stack.append(input_q.pop())
                else:
                    break

        # Extract result — should be END + result on stack
        # Find the result (skip END markers)
        result_node = None
        for cat, node in stack:
            if cat != CAT_END and node is not None:
                result_node = node
        if result_node is None:
            return Num(0)
        return result_node

    def _current(self) -> Token:
        return self._tokens[self._pos]

    def _peek(self) -> Token | None:
        if self._pos + 1 < len(self._tokens):
            return self._tokens[self._pos + 1]
        return None

    def _eat(self, token_type: str) -> Token:
        token = self._current()
        if token.type != token_type:
            raise SyntaxError_(f"Expected {token_type}, got {token.type}")
        self._pos += 1
        return token

    def _parse_dfn(self) -> Dfn:
        """Parse a dfn: { statement (⋄ statement)* }"""
        self._eat(TokenType.LBRACE)
        statements: list[object] = []
        while self._current().type != TokenType.RBRACE:
            stmt = self._parse_dfn_statement()
            statements.append(stmt)
            if self._current().type == TokenType.DIAMOND:
                self._eat(TokenType.DIAMOND)
        self._eat(TokenType.RBRACE)
        return Dfn(statements)

    def _parse_dfn_statement(self) -> object:
        """Parse a statement inside a dfn, handling guards and ⍺← default."""
        # Check for ⍺← default
        peek = self._peek()
        if (
            self._current().type == TokenType.ALPHA
            and peek is not None
            and peek.type == TokenType.ASSIGN
        ):
            self._eat(TokenType.ALPHA)
            self._eat(TokenType.ASSIGN)
            default = self._parse_statement()
            return AlphaDefault(default)

        stmt = self._parse_statement()
        # Check for guard: expr : expr
        if self._current().type == TokenType.GUARD:
            self._eat(TokenType.GUARD)
            body = self._parse_statement()
            return Guard(stmt, body)
        return stmt

    def _parse_atom(self) -> object:
        token = self._current()
        if token.type == TokenType.LPAREN:
            self._eat(TokenType.LPAREN)
            # Check for bare function glyph as operand: (-)  (+)  (⍳)  etc.
            if (
                self._current().type == TokenType.FUNCTION
                and self._pos + 1 < len(self._tokens)
                and self._tokens[self._pos + 1].type == TokenType.RPAREN
            ):
                fn_token = self._eat(TokenType.FUNCTION)
                self._eat(TokenType.RPAREN)
                assert isinstance(fn_token.value, str)
                return FunctionRef(fn_token.value)
            result = self._parse_statement()
            self._eat(TokenType.RPAREN)
            return result
        if token.type == TokenType.NUMBER:
            self._eat(TokenType.NUMBER)
            assert isinstance(token.value, (int, float))
            return Num(token.value)
        if token.type == TokenType.STRING:
            self._eat(TokenType.STRING)
            assert isinstance(token.value, str)
            return Str(token.value)
        if token.type == TokenType.ID:
            self._eat(TokenType.ID)
            assert isinstance(token.value, str)
            return Var(token.value)
        if token.type == TokenType.QUALIFIED_NAME:
            self._eat(TokenType.QUALIFIED_NAME)
            assert isinstance(token.value, str)
            return QualifiedVar(token.value.split("::"))
        if token.type == TokenType.OMEGA:
            self._eat(TokenType.OMEGA)
            return Omega()
        if token.type == TokenType.ALPHA:
            self._eat(TokenType.ALPHA)
            return Alpha()
        if token.type == TokenType.ALPHA_ALPHA:
            self._eat(TokenType.ALPHA_ALPHA)
            return AlphaAlpha()
        if token.type == TokenType.OMEGA_OMEGA:
            self._eat(TokenType.OMEGA_OMEGA)
            return OmegaOmega()
        if token.type == TokenType.NABLA:
            self._eat(TokenType.NABLA)
            return Nabla()
        if token.type == TokenType.SYSVAR:
            self._eat(TokenType.SYSVAR)
            assert isinstance(token.value, str)
            return SysVar(token.value)
        if token.type == TokenType.LBRACE:
            return self._parse_dfn()
        raise SyntaxError_(f"Unexpected token: {token}")

    def _parse_atom_with_index(self) -> object:
        """Parse an atom, then check for bracket indexing."""
        atom = self._parse_atom()
        if self._current().type == TokenType.LBRACKET:
            return self._parse_bracket_index(atom)
        return atom

    def _parse_bracket_index(self, array: object) -> Index:
        """Parse [idx] or [row;col] bracket indexing."""
        self._eat(TokenType.LBRACKET)
        indices: list[object | None] = []
        # First index (may be empty for [;col])
        if self._current().type == TokenType.SEMICOLON:
            indices.append(None)
        elif self._current().type == TokenType.RBRACKET:
            indices.append(None)
        else:
            indices.append(self._parse_statement())
        # Additional indices separated by ;
        while self._current().type == TokenType.SEMICOLON:
            self._eat(TokenType.SEMICOLON)
            if self._current().type in (TokenType.SEMICOLON, TokenType.RBRACKET):
                indices.append(None)
            else:
                indices.append(self._parse_statement())
        self._eat(TokenType.RBRACKET)
        return Index(array, indices)

    def _is_array_start(self) -> bool:
        return self._current().type in (
            TokenType.NUMBER, TokenType.LPAREN, TokenType.ID,
            TokenType.OMEGA, TokenType.ALPHA, TokenType.ALPHA_ALPHA,
            TokenType.OMEGA_OMEGA, TokenType.NABLA,
            TokenType.LBRACE, TokenType.STRING, TokenType.QUALIFIED_NAME,
            TokenType.SYSVAR,
        )

    def _parse_array(self) -> object:
        """Parse one or more adjacent numeric atoms as a vector,
        or a single non-numeric atom."""
        first = self._parse_atom_with_index()
        if not isinstance(first, Num):
            return first
        elements: list[Num] = [first]
        while self._current().type == TokenType.NUMBER:
            token = self._eat(TokenType.NUMBER)
            assert isinstance(token.value, (int, float))
            elements.append(Num(token.value))
        if len(elements) == 1:
            return elements[0]
        return Vector(elements)

    def _parse_function_expr(self) -> tuple[str, str | None]:
        """Parse a function glyph, possibly followed by an operator.
        Returns (function_glyph, operator_glyph_or_None)."""
        func_token = self._eat(TokenType.FUNCTION)
        assert isinstance(func_token.value, str)
        if self._current().type == TokenType.OPERATOR:
            op_token = self._eat(TokenType.OPERATOR)
            assert isinstance(op_token.value, str)
            return func_token.value, op_token.value
        return func_token.value, None

    def _parse_statement(self) -> object:
        """Parse a statement using Iverson's stack-based algorithm."""
        items = self._build_items()
        if not items:
            return Num(0)
        result = self._stack_parse(items)
        # Handle trailing bracket indexing: result[idx]
        if (self._pos < len(self._tokens)
                and self._current().type == TokenType.LBRACKET):
            result = self._parse_bracket_index(result)
        return result

    def parse(self) -> object:
        # Empty input (e.g. comment-only line) → return Num(0) as no-op
        if self._current().type == TokenType.EOF:
            return Num(0)
        statements = [self._parse_statement()]
        while self._current().type == TokenType.DIAMOND:
            self._eat(TokenType.DIAMOND)
            statements.append(self._parse_statement())
        if self._current().type != TokenType.EOF:
            raise SyntaxError_(f"Unexpected token after expression: {self._current()}")
        if len(statements) == 1:
            return statements[0]
        return Program(statements)


def parse(source: str, name_table: dict[str, int] | None = None,
          operator_arity: dict[str, int] | None = None) -> object:
    tokens = Tokenizer(source).tokenize()
    return Parser(tokens, name_table, operator_arity).parse()


def is_balanced(source: str) -> bool:
    """Check if braces and brackets are balanced.

    Returns True if the source has matching { } pairs.
    Returns False if there are unclosed openers (incomplete input).
    """
    depth = 0
    in_string = False
    for ch in source:
        if ch == "'" and not in_string:
            in_string = True
        elif ch == "'" and in_string:
            in_string = False
        elif not in_string:
            if ch == "{":
                depth += 1
            elif ch == "}":
                depth -= 1
    return depth == 0



"""Tests that pico_eval.py imports and API still work."""

import marple
from marple.engine import Interpreter


def test_no_dataclasses_in_engine() -> None:
    """engine.py must not use dataclasses — MicroPython doesn't have them."""
    import ast
    with open("src/marple/engine.py") as f:
        source = f.read()
    tree = ast.parse(source)
    for node in ast.walk(tree):
        if isinstance(node, ast.ImportFrom) and node.module == "dataclasses":
            raise AssertionError("engine.py imports dataclasses — breaks MicroPython")


def test_no_importlib_in_executor() -> None:
    """executor.py must not use importlib — MicroPython doesn't have it."""
    import ast
    with open("src/marple/executor.py") as f:
        source = f.read()
    tree = ast.parse(source)
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                if alias.name == "importlib":
                    raise AssertionError("executor.py imports importlib — breaks MicroPython")


def test_no_os_path_in_os_filesystem() -> None:
    """os_filesystem.py must not use os.path — MicroPython doesn't have it."""
    import ast
    with open("src/marple/adapters/os_filesystem.py") as f:
        source = f.read()
    tree = ast.parse(source)
    for node in ast.walk(tree):
        if isinstance(node, ast.Attribute) and isinstance(node.value, ast.Attribute):
            if (isinstance(node.value.value, ast.Name) and
                    node.value.value.id == "os" and node.value.attr == "path"):
                raise AssertionError(
                    f"os_filesystem.py uses os.path.{node.attr} — breaks MicroPython")


def test_pico_eval_imports() -> None:
    """Simulate the imports pico_eval.py does on startup."""
    from marple.engine import Interpreter  # noqa: F811
    from marple.errors import APLError  # noqa: F401
    import marple as m  # noqa: F401


def test_pico_eval_create_interpreter() -> None:
    """pico_eval.py creates Interpreter() with no arguments."""
    interp = Interpreter()
    assert interp is not None


def test_pico_eval_version_accessible() -> None:
    """pico_eval.py reads marple.__version__ for the banner."""
    assert hasattr(marple, "__version__")
    assert isinstance(marple.__version__, str)


def test_pico_eval_execute_api() -> None:
    """pico_eval.py uses interp.execute() which returns EvalResult."""
    interp = Interpreter()
    r = interp.execute("2+3")
    assert not r.silent
    assert r.display_text == "5"


def test_pico_eval_silent_assignment() -> None:
    """Assignments are silent — pico_eval.py checks r.silent."""
    interp = Interpreter()
    r = interp.execute("x←42")
    assert r.silent


def test_pico_eval_full_startup_sequence() -> None:
    """Simulate the exact startup sequence from pico_eval.py."""
    # Line 31-33: imports
    from marple.engine import Interpreter as Interp  # noqa: F811
    from marple.errors import APLError  # noqa: F401
    import marple as m

    # Line 37: create interpreter
    interp = Interp()

    # Line 43: version string for banner
    banner = "MARPLE v" + m.__version__
    assert "MARPLE v" in banner

    # Lines 65-71: execute and check result
    r = interp.execute("1+1")
    if r.silent:
        output = ""
    else:
        output = r.display_text
    assert output == "2"



"""User configuration for MARPLE.

Reads ~/.marple/config.ini if it exists. Example:

    [defaults]
    io = 0

    [paths]
    workspaces = ~/marple-workspaces
    sessions = ~/marple-sessions
"""
import os

try:
    from configparser import ConfigParser
    _CONFIG_DIR = os.path.expanduser("~/.marple")
    _CONFIG_PATH = os.path.join(_CONFIG_DIR, "config.ini")
    _config: object = ConfigParser()
    if os.path.exists(_CONFIG_PATH):
        _config.read(_CONFIG_PATH)  # type: ignore[union-attr]
except (ImportError, AttributeError):
    _config = None


def get_default_io() -> int:
    """Default ⎕IO value (0 or 1)."""
    if _config is None:
        return 1
    return _config.getint("defaults", "io", fallback=1)  # type: ignore[union-attr]


def get_workspaces_dir() -> str:
    """Root directory for workspace saves."""
    if _config is None:
        return "workspaces"
    path = _config.get("paths", "workspaces", fallback="workspaces")  # type: ignore[union-attr]
    return os.path.expanduser(path)


def get_sessions_dir() -> str:
    """Directory for session saves."""
    if _config is None:
        return "sessions"
    path = _config.get("paths", "sessions", fallback="sessions")  # type: ignore[union-attr]
    return os.path.expanduser(path)



"""Shared formatting for MARPLE display and ⍕."""

try:
    from typing import Any
except ImportError:
    pass

from marple.arraymodel import APLArray


def format_num(x: Any, pp: int = 10) -> str:
    """Format a number for display, using pp significant digits for floats."""
    # Convert numpy scalars to Python types
    if hasattr(x, "item"):
        x = x.item()  # type: ignore[union-attr]
    if isinstance(x, bool):
        return str(int(x))
    if isinstance(x, float):
        if x == int(x) and abs(x) < 1e15:
            n = int(x)
            return "¯" + str(abs(n)) if n < 0 else str(n)
        s = f"{x:.{pp}g}"
        if s.startswith("-"):
            s = "¯" + s[1:]
        return s
    if isinstance(x, int) and x < 0:
        return "¯" + str(abs(x))
    # Decimal support
    try:
        from decimal import Decimal
        if isinstance(x, Decimal):
            s = str(x)
            if s.startswith("-"):
                return "¯" + s[1:]
            return s
    except ImportError:
        pass
    return str(x)


def _rjust(s: str, width: int) -> str:
    if len(s) >= width:
        return s
    return " " * (width - len(s)) + s


def _is_char_array(arr: APLArray) -> bool:
    return len(arr.data) > 0 and all(isinstance(x, str) for x in arr.data)


def _format_matrix(result: APLArray, pp: int) -> str:
    """Format a rank-2 array as right-justified columns."""
    rows, cols = result.shape
    if _is_char_array(result):
        lines = []
        for r in range(rows):
            row_data = result.data[r * cols:(r + 1) * cols]
            lines.append("".join(str(x) for x in row_data))
        return "\n".join(lines)
    strs = [format_num(result.data[r * cols + c], pp) for r in range(rows) for c in range(cols)]
    col_widths = []
    for c in range(cols):
        w = max(len(strs[r * cols + c]) for r in range(rows))
        col_widths.append(w)
    lines = []
    for r in range(rows):
        parts = []
        for c in range(cols):
            parts.append(_rjust(strs[r * cols + c], col_widths[c]))
        lines.append(" ".join(parts))
    return "\n".join(lines)


def format_result(result: APLArray, env: Any = None) -> str:
    """Format an APLArray for display."""
    pp = 10
    if env is not None:
        pp_val = env.get("⎕PP")
        if pp_val is not None:
            pp = int(pp_val.data[0])
    if result.is_scalar():
        return format_num(result.data[0], pp)
    if _is_char_array(result):
        if len(result.shape) == 1:
            return "".join(str(x) for x in result.data)
        if len(result.shape) == 2:
            return _format_matrix(result, pp)
    if len(result.shape) == 1:
        return " ".join(format_num(x, pp) for x in result.data)
    if len(result.shape) == 2:
        return _format_matrix(result, pp)
    # Rank 3+: format each 2D slice separated by blank lines
    if len(result.shape) >= 3:
        slice_size = result.shape[-2] * result.shape[-1]
        num_slices = len(result.data) // slice_size
        slices = []
        for s in range(num_slices):
            start = s * slice_size
            slice_data = list(result.data[start:start + slice_size])
            slice_arr = APLArray([result.shape[-2], result.shape[-1]], slice_data)
            slices.append(_format_matrix(slice_arr, pp))
        return "\n\n".join(slices)
    return repr(result)



"""Executor: shared AST evaluation logic for the MARPLE interpreter."""

from __future__ import annotations

import time

_START_TIME = time.time()

from typing import Any, TYPE_CHECKING

from marple.arraymodel import APLArray, S
from marple.backend import (
    _DOWNCAST_CT, is_numeric_array, maybe_downcast,
)
from marple.cells import clamp_rank, decompose, reassemble, resolve_rank_spec
from marple.errors import DomainError, SyntaxError_, ValueError_
from marple.dyadic_functions import DyadicFunctionBinding
from marple.monadic_functions import MonadicFunctionBinding
from marple.operator_binding import DerivedFunctionBinding
from marple.parser import (
    BoundOperator,
    Dfn,
    FunctionRef,
    Node,
    RankDerived,
    ReduceOp,
    ScanOp,
)
from marple.symbol_table import NC_ARRAY, NC_FUNCTION, NC_OPERATOR, NC_UNKNOWN

if TYPE_CHECKING:
    from marple.environment import Environment
    from marple.ports.filesystem import FileSystem

_READONLY_QUADS = frozenset({"⎕A", "⎕D", "⎕TS", "⎕EN", "⎕DM"})


def _newlines_to_diamonds(source: str) -> str:
    """Convert newlines to ⋄ (statement separator), preserving strings."""
    result: list[str] = []
    in_string = False
    for ch in source:
        if ch == "'" and not in_string:
            in_string = True
            result.append(ch)
        elif ch == "'" and in_string:
            in_string = False
            result.append(ch)
        elif ch == "\n" and not in_string:
            result.append("⋄")
        else:
            result.append(ch)
    return "".join(result)


def _ljust(s: str, width: int) -> str:
    return s + " " * max(0, width - len(s))


def _apl_chars_to_str(data: Any) -> str:
    """Convert an APLArray's character data to a Python string."""
    return "".join(str(c) for c in data)


def _name_class(value: object) -> int:
    """Return the APL name class for a value."""
    from marple.dfn_binding import DfnBinding
    if isinstance(value, DfnBinding):
        return NC_FUNCTION
    if isinstance(value, APLArray):
        return NC_ARRAY
    return NC_UNKNOWN


class Executor:
    """Base class providing AST evaluation, shared by Interpreter and DfnBinding."""

    env: Environment

    @property
    def fs(self) -> 'FileSystem':
        return self.env.fs

    # ── String-keyed dispatch tables (class-level, shared) ──

    _SYSVAR_DISPATCH: dict[str, str] = {
        "⎕TS": "_sysvar_ts",
        "⎕AI": "_sysvar_ai",
        "⎕VER": "_sysvar_ver",
        "⍞": "_sysvar_quote_quad",
        "⎕": "_sysvar_quad",
    }

    _SYS_FN_DISPATCH: dict[str, str] = {
        "⎕NC": "_sys_nc",
        "⎕EX": "_sys_ex",
        "⎕NL": "_sys_nl",
        "⎕UCS": "_sys_ucs",
        "⎕DR": "_sys_dr",
        "⎕SIGNAL": "_sys_signal",
        "⎕NREAD": "_sys_nread",
        "⎕NEXISTS": "_sys_nexists",
        "⎕NDELETE": "_sys_ndelete",
        "⎕CR": "_sys_cr",
        "⎕FX": "_sys_fx",
        "⎕DL": "_sys_dl",
        "⎕CSV": "_sys_csv",
    }

    # ── Core evaluation ──

    def evaluate(self, node: Node | object) -> APLArray:
        """Evaluate an AST node by calling its execute method."""
        if isinstance(node, Node):
            return node.execute(self)  # type: ignore[return-value]
        raise DomainError(f"Unknown AST node: {type(node)}")

    # ── Callback methods for node execute() ──

    def dispatch_monadic(self, glyph: str, operand: APLArray) -> APLArray:
        if glyph == "⍎":
            return self._execute_string(operand)
        return MonadicFunctionBinding(self.env).apply(glyph, operand)

    def _execute_string(self, operand: APLArray) -> APLArray:
        from marple.parser import parse
        source = "".join(str(c) for c in operand.data)
        tree = parse(source, self.env.class_dict())
        return self.evaluate(tree)

    def dispatch_dyadic(self, glyph: str, left: APLArray, right: APLArray) -> APLArray:
        return DyadicFunctionBinding(self.env).apply(glyph, left, right)

    def resolve_dyadic(self, fn: object) -> object:
        """Resolve a function reference to a dyadic callable."""
        from marple.parser import FunctionRef
        glyph = fn.glyph if isinstance(fn, FunctionRef) else fn
        if not isinstance(glyph, str):
            raise DomainError(f"Expected function for operator, got {type(fn)}")
        binding = DyadicFunctionBinding(self.env)
        return binding.resolve_with_env(glyph)

    def call_ibeam(self, path: str, operand: APLArray) -> APLArray:
        """Call a Python function via i-beam."""
        parts = path.rsplit(".", 1)
        if len(parts) != 2:
            raise DomainError(f"Invalid i-beam path: {path}")
        module_name, func_name = parts
        try:
            mod = __import__(module_name)
            for part in module_name.split(".")[1:]:
                mod = getattr(mod, part)
        except ImportError:
            raise DomainError(f"Cannot import module: {module_name}")
        func = getattr(mod, func_name, None)
        if func is None:
            raise DomainError(f"Function not found: {path}")
        result = func(operand)
        if not isinstance(result, APLArray):
            raise DomainError(f"I-beam function must return APLArray: {path}")
        return result

    def resolve_qualified(self, parts: list[str]) -> object:
        from marple.namespace import Namespace, load_system_workspace
        if parts[0] == "$":
            import marple.stdlib
            f = marple.stdlib.__file__
            stdlib_path = f[:f.rfind("/")] if "/" in f else f[:f.rfind("\\")]
            sys_ws = load_system_workspace(stdlib_path)
            result = sys_ws.resolve(parts[1:])
            if result is None:
                raise DomainError("Undefined: " + "::".join(parts))
            return result
        raise DomainError(f"Undefined namespace: {parts[0]}")

    def apply_derived(self, operator: str, function: object, operand: APLArray) -> APLArray:
        # If function is an AST node (e.g. Dfn), evaluate it first
        from marple.dfn_binding import DfnBinding
        from marple.parser import Node as ParserNode
        if isinstance(function, ParserNode):
            val = self.evaluate(function)
            if isinstance(val, DfnBinding):
                function = lambda a, o, _b=val: _b.apply(o, alpha=a)
            elif isinstance(val, FunctionRef):
                function = val.glyph
        return DerivedFunctionBinding().apply(operator, function, operand)

    def assign(self, name: str, value_node: object) -> APLArray:
        if name in _READONLY_QUADS:
            raise DomainError(f"Cannot assign to read-only system variable {name}")
        value = self.evaluate(value_node)
        if name in ("⎕", "⍞"):
            return self._io_assign(name, value)
        if isinstance(value, APLArray) and is_numeric_array(value.data):
            value = APLArray(list(value.shape), maybe_downcast(value.data, _DOWNCAST_CT))
        if name.startswith("⎕"):
            if name == "⎕FR" and isinstance(value, APLArray):
                fr_val = int(value.data[0])
                if fr_val not in (645, 1287):
                    raise DomainError(f"⎕FR must be 645 or 1287, got {fr_val}")
            if name == "⎕RL" and isinstance(value, APLArray):
                import random as _random
                _random.seed(int(value.data[0]))
            self.env[name] = value
        else:
            new_class = _name_class(value)
            old_class = self.env.name_class(name)
            if old_class != 0 and new_class != 0 and old_class != new_class:
                from marple.errors import ClassError
                raise ClassError(f"Cannot change class of '{name}' from {old_class} to {new_class}")
            self.env.bind_name(name, value, new_class)
        return value if isinstance(value, APLArray) else S(0)

    def _io_assign(self, name: str, value: APLArray) -> APLArray:
        """Handle ⎕← (output with newline) and ⍞← (prompt, read, return prompt+response)."""
        if self.env.console is None:
            raise DomainError("Console not available for I/O")
        from marple.formatting import format_result
        text = format_result(value, self.env)
        if name == "⎕":
            self.env.console.writeln(text)
            return value
        # ⍞← : display prompt, read input, return response only (Dyalog style)
        line = self.env.console.read_line(text)
        if line is None:
            raise DomainError("⍞ input not available — use the terminal REPL for interactive input")
        self.env.console.writeln(text + line)
        response_chars: list[object] = list(line)
        return APLArray([len(response_chars)], response_chars)

    def create_binding(self, dfn_node: Dfn) -> object:
        from marple.dfn_binding import DfnBinding
        # Store a reference to env, not a copy — names added later
        # (e.g. forward references) are visible at call time when
        # DfnBinding._make_env copies the env.
        return DfnBinding(dfn_node, self.env)

    def eval_sysvar(self, name: str) -> APLArray:
        method_name = self._SYSVAR_DISPATCH.get(name)
        if method_name is not None:
            return getattr(self, method_name)()
        if name not in self.env:
            if name in self._SYS_FN_DISPATCH or name in self._DYADIC_SYS_FN_DISPATCH:
                raise SyntaxError_(f"{name} is a system function — it requires an argument")
            raise ValueError_(f"Undefined system variable: {name}")
        return self.env[name]

    # ── System variables ──

    def _sysvar_ts(self) -> APLArray:
        now = time.time()
        t = time.localtime(now)
        ms = int((now % 1) * 1000)
        return APLArray([7], [t[0], t[1], t[2], t[3], t[4], t[5], ms])

    def _sysvar_ai(self) -> APLArray:
        # 1: user ID
        try:
            import os
            uid = os.getuid()
        except AttributeError:
            uid = 1000  # MicroPython
        # 2: compute time (CPU ms)
        try:
            import resource
            cpu_ms = int(resource.getrusage(resource.RUSAGE_SELF).ru_utime * 1000)
        except (ImportError, AttributeError):
            cpu_ms = 0  # MicroPython
        # 3: connect time (ms since session start)
        if hasattr(time, "ticks_ms"):
            elapsed_ms = time.ticks_ms()  # type: ignore[attr-defined]
        else:
            elapsed_ms = int((time.time() - _START_TIME) * 1000)
        # 4: keying time (not tracked)
        return APLArray([4], [uid, cpu_ms, elapsed_ms, 0])

    def _sysvar_ver(self) -> APLArray:
        from marple import __version__
        import sys
        s = "MARPLE v" + __version__ + " on " + sys.platform
        return APLArray([len(s)], list(s))

    def _sysvar_quad(self) -> APLArray:
        """⎕ — prompt, read, parse, and evaluate input as APL."""
        if self.env.console is None:
            raise DomainError("Console not available for I/O")
        line = self.env.console.read_line("⎕:")
        if line is None:
            raise DomainError("⎕ input not available — use the terminal REPL for interactive input")
        self.env.console.writeln("⎕:" + line)
        from marple.parser import parse
        tree = parse(line, self.env.class_dict())
        return self.evaluate(tree)

    def _sysvar_quote_quad(self) -> APLArray:
        """⍞ — read a line of character input (no prompt)."""
        if self.env.console is None:
            raise DomainError("Console not available for I/O")
        line = self.env.console.read_line("")
        if line is None:
            raise DomainError("⍞ input not available — use the terminal REPL for interactive input")
        self.env.console.writeln(line)
        return APLArray([len(line)], list(line))

    # ── Rank operator ──

    def apply_rank_monadic(self, rank_node: RankDerived, operand_node: object) -> APLArray:
        omega = self.evaluate(operand_node)
        rank_spec_val = self.evaluate(rank_node.rank_spec)
        a, _, _ = resolve_rank_spec(rank_spec_val)
        k = clamp_rank(a, len(omega.shape))
        frame_shape, cells = decompose(omega, k)
        results = [self.apply_func_monadic(rank_node.function, cell) for cell in cells]
        return reassemble(frame_shape, results)

    def apply_func_monadic(self, func: object, omega: APLArray) -> APLArray:
        """Apply a function monadically. Used by rank operator."""
        if isinstance(func, str):
            return MonadicFunctionBinding(self.env).apply(func, omega)
        if isinstance(func, FunctionRef):
            return MonadicFunctionBinding(self.env).apply(func.glyph, omega)
        if isinstance(func, ReduceOp):
            return DerivedFunctionBinding().apply("/", func.function, omega)
        if isinstance(func, ScanOp):
            return DerivedFunctionBinding().apply("\\", func.function, omega)
        if isinstance(func, BoundOperator) and isinstance(func.operator, str):
            return DerivedFunctionBinding().apply(
                func.operator, func.left_operand, omega)
        # AST node (Var, Dfn, etc.) — evaluate to get the function value, then apply
        from marple.dfn_binding import DfnBinding
        from marple.parser import Node
        if isinstance(func, Node):
            val = self.evaluate(func)
            if isinstance(val, DfnBinding):
                return val.apply(omega)
            if isinstance(val, FunctionRef):
                return MonadicFunctionBinding(self.env).apply(val.glyph, omega)
            if hasattr(val, 'dfn') and hasattr(val, 'env'):
                binding = DfnBinding(getattr(val, 'dfn'), self.env)
                return binding.apply(omega)
        raise DomainError(f"Expected function for rank, got {type(func)}")

    def apply_rank_dyadic(self, rank_node: object, left_node: object, right_node: object) -> APLArray:
        from marple.errors import LengthError
        from marple.parser import RankDerived
        assert isinstance(rank_node, RankDerived)
        alpha = self.evaluate(left_node)
        omega = self.evaluate(right_node)
        rank_spec_val = self.evaluate(rank_node.rank_spec)
        _, b_rank, c_rank = resolve_rank_spec(rank_spec_val)
        b = clamp_rank(b_rank, len(alpha.shape))
        c = clamp_rank(c_rank, len(omega.shape))
        left_frame, left_cells = decompose(alpha, b)
        right_frame, right_cells = decompose(omega, c)
        if left_frame == right_frame:
            pairs = list(zip(left_cells, right_cells))
            frame = left_frame
        elif left_frame == []:
            pairs = [(left_cells[0], rc) for rc in right_cells]
            frame = right_frame
        elif right_frame == []:
            pairs = [(lc, right_cells[0]) for lc in left_cells]
            frame = left_frame
        else:
            raise LengthError(f"Frame mismatch: {left_frame} vs {right_frame}")
        results = [self._apply_func_dyadic(rank_node.function, lc, rc) for lc, rc in pairs]
        return reassemble(frame, results)

    def _apply_func_dyadic(self, func: object, alpha: APLArray, omega: APLArray) -> APLArray:
        """Apply a function dyadically. Used by dyadic rank operator."""
        from marple.parser import FunctionRef
        if isinstance(func, str):
            return DyadicFunctionBinding(self.env).apply(func, alpha, omega)
        if isinstance(func, FunctionRef):
            return DyadicFunctionBinding(self.env).apply(func.glyph, alpha, omega)
        from marple.dfn_binding import DfnBinding
        from marple.parser import Node
        if isinstance(func, Node):
            val = self.evaluate(func)
            if isinstance(val, DfnBinding):
                return val.apply(omega, alpha=alpha)
            if isinstance(val, FunctionRef):
                return DyadicFunctionBinding(self.env).apply(val.glyph, alpha, omega)
        raise DomainError(f"Expected function for rank, got {type(func)}")

    # ── Power operator ──

    def apply_power_monadic(self, power_node: object, operand_node: object) -> APLArray:
        from marple.dfn_binding import DfnBinding
        from marple.parser import PowerDerived
        assert isinstance(power_node, PowerDerived)
        omega = self.evaluate(operand_node)
        right_op = power_node.right_operand
        right_val = self._resolve_power_operand(right_op)
        if isinstance(right_val, APLArray) and right_val.is_scalar():
            n = int(right_val.data[0])
            if n < 0:
                raise DomainError("DOMAIN ERROR: inverse (⍣ with negative) not supported")
            result = omega
            for _ in range(n):
                result = self.apply_func_monadic(power_node.function, result)
            return result
        if isinstance(right_val, (DfnBinding, FunctionRef)):
            prev = omega
            while True:
                curr = self.apply_func_monadic(power_node.function, prev)
                test = self._apply_func_dyadic_or_match(right_val, curr, prev)
                if test.data[0]:
                    return curr
                prev = curr
        raise DomainError("⍣ right operand must be integer or function")

    def apply_power_dyadic(self, power_node: object, left_node: object,
                           right_node: object) -> APLArray:
        from marple.dfn_binding import DfnBinding
        from marple.parser import PowerDerived
        assert isinstance(power_node, PowerDerived)
        alpha = self.evaluate(left_node)
        omega = self.evaluate(right_node)
        right_op = power_node.right_operand
        right_val = self._resolve_power_operand(right_op)
        if isinstance(right_val, APLArray) and right_val.is_scalar():
            n = int(right_val.data[0])
            if n < 0:
                raise DomainError("DOMAIN ERROR: inverse (⍣ with negative) not supported")
            result = omega
            for _ in range(n):
                result = self._apply_func_dyadic(power_node.function, alpha, result)
            return result
        if isinstance(right_val, (DfnBinding, FunctionRef)):
            prev = omega
            while True:
                curr = self._apply_func_dyadic(power_node.function, alpha, prev)
                test = self._apply_func_dyadic_or_match(right_val, curr, prev)
                if test.data[0]:
                    return curr
                prev = curr
        raise DomainError("⍣ right operand must be integer or function")

    def _resolve_power_operand(self, right_op: object) -> object:
        """Resolve the right operand of ⍣ — may be a glyph string, AST node, or value."""
        from marple.dfn_binding import DfnBinding
        if isinstance(right_op, str):
            # Primitive glyph like ≡ or =
            return FunctionRef(right_op)
        if isinstance(right_op, FunctionRef):
            return right_op
        result = self.evaluate(right_op)
        if isinstance(result, (DfnBinding, FunctionRef)):
            return result
        return result

    def _apply_func_dyadic_or_match(self, func: object, left: APLArray,
                                     right: APLArray) -> APLArray:
        """Apply a dyadic function for convergence test. Handles FunctionRef for ≡ and =."""
        from marple.dfn_binding import DfnBinding
        if isinstance(func, FunctionRef):
            if func.glyph == "≡":
                return S(1 if left == right else 0)
            if func.glyph == "=":
                return S(1 if left.data[0] == right.data[0] else 0)
            return DyadicFunctionBinding(self.env).apply(func.glyph, left, right)
        if isinstance(func, DfnBinding):
            return func.apply(right, alpha=left)
        raise DomainError("⍣ convergence function must be a function")

    # ── System functions ──

    _DYADIC_SYS_FN_DISPATCH: dict[str, str] = {
        "⎕EA": "_sys_ea",
        "⎕DR": "_sys_dr_dyadic",
        "⎕NWRITE": "_sys_nwrite",
    }

    def dispatch_sys_monadic(self, name: str, operand_node: object) -> APLArray:
        if name == "⎕FMT":
            return self._sys_fmt_monadic(operand_node)
        operand = self.evaluate(operand_node)
        method_name = self._SYS_FN_DISPATCH.get(name)
        if method_name is not None:
            return getattr(self, method_name)(operand)
        raise DomainError(f"Unknown system function: {name}")

    def dispatch_sys_dyadic(self, name: str, left_node: object, right_node: object) -> APLArray:
        if name == "⎕FMT":
            return self._sys_fmt_dyadic(left_node, right_node)
        method_name = self._DYADIC_SYS_FN_DISPATCH.get(name)
        if method_name is not None:
            left = self.evaluate(left_node)
            right = self.evaluate(right_node)
            return getattr(self, method_name)(left, right)
        raise DomainError(f"Unknown dyadic system function: {name}")

    def _sys_nc(self, operand: APLArray) -> APLArray:
        return S(self.env.name_class(_apl_chars_to_str(operand.data)))

    def _sys_ex(self, operand: APLArray) -> APLArray:
        if len(operand.shape) == 2:
            return self._sys_ex_matrix(operand)
        return self._expunge_name(_apl_chars_to_str(operand.data).rstrip())

    def _sys_ex_matrix(self, operand: APLArray) -> APLArray:
        rows, cols = operand.shape
        count = 0
        for r in range(rows):
            start = r * cols
            name = _apl_chars_to_str(operand.data[start:start + cols]).rstrip()
            result = self._expunge_name(name)
            count += int(result.data[0])
        return S(count)

    def _expunge_name(self, name: str) -> APLArray:
        """Remove a single name from the symbol table."""
        return S(1) if self.env.delete_name(name) else S(0)

    def _sys_nl(self, operand: APLArray) -> APLArray:
        nc = int(operand.data[0])
        names = self.env.names_of_class(nc)
        if not names:
            return APLArray([0, 0], [])
        max_len = max(len(n) for n in names)
        chars: list[object] = []
        for n in names:
            chars.extend(list(_ljust(n, max_len)))
        return APLArray([len(names), max_len], chars)

    def _sys_ucs(self, operand: APLArray) -> APLArray:
        from marple.backend import to_list
        if all(isinstance(x, str) for x in operand.data):
            return APLArray(list(operand.shape), [ord(c) for c in operand.data])
        data = to_list(operand.data)
        if operand.is_scalar():
            return APLArray([], [chr(int(data[0]))])
        return APLArray(list(operand.shape), [chr(int(x)) for x in data])

    def _sys_dr(self, operand: APLArray) -> APLArray:
        from marple.backend import is_numeric_array, to_list, _is_float_dtype
        if len(operand.data) > 0 and isinstance(to_list(operand.data)[0], str):
            return S(80)
        if is_numeric_array(operand.data):
            dtype_str = str(operand.data.dtype)
            if "uint8" in dtype_str:
                return S(11)
            if "int8" in dtype_str:
                return S(83)
            if "int16" in dtype_str:
                return S(163)
            if "int32" in dtype_str:
                return S(323)
            if "int64" in dtype_str:
                return S(645)
            if _is_float_dtype(operand.data):
                return S(645)
        vals = to_list(operand.data)
        if vals:
            v = vals[0]
            if isinstance(v, str):
                return S(80)
            if isinstance(v, float):
                return S(645)
            if isinstance(v, int):
                return S(323)
        return S(0)

    def _sys_signal(self, operand: APLArray) -> APLArray:
        from marple.errors import (
            SyntaxError_, ValueError_, LengthError,
            RankError, IndexError_, LimitError, SecurityError,
        )
        code = int(operand.data[0])
        error_map: dict[int, type] = {
            1: SyntaxError_, 2: ValueError_, 3: DomainError, 4: LengthError,
            5: RankError, 6: IndexError_, 7: LimitError, 9: SecurityError,
        }
        err_class = error_map.get(code, DomainError)
        raise err_class(f"Signalled by ⎕SIGNAL {code}")

    def _sys_ea(self, left: APLArray, right: APLArray) -> APLArray:
        """⎕EA: error-guarded execution. Try right; on error, execute left."""
        from marple.errors import APLError
        from marple.parser import parse
        right_str = _apl_chars_to_str(right.data)
        try:
            tree = parse(right_str, self.env.class_dict())
            return self.evaluate(tree)
        except APLError as e:
            self.env["⎕EN"] = S(e.code)
            self.env["⎕DM"] = APLArray([len(str(e))], list(str(e)))
            left_str = _apl_chars_to_str(left.data)
            tree = parse(left_str, self.env.class_dict())
            return self.evaluate(tree)

    def _sys_dr_dyadic(self, left: APLArray, right: APLArray) -> APLArray:
        """Dyadic ⎕DR: convert data representation."""
        from marple.backend import to_list
        target = int(left.data[0])
        vals = to_list(right.data)
        if target == 645:
            new_data = [float(v) for v in vals]
            return APLArray(list(right.shape), new_data)
        if target in (323, 163, 83):
            new_data = [int(round(v)) for v in vals]
            return APLArray(list(right.shape), new_data)
        if target == 11:
            from marple.backend import to_bool_array
            new_data = to_bool_array([int(bool(v)) for v in vals])
            return APLArray(list(right.shape), new_data)
        if target == 80:
            new_data = [chr(int(v)) for v in vals]
            return APLArray(list(right.shape), new_data)
        raise DomainError("Invalid ⎕DR type code: " + str(target))

    def _sys_fmt_monadic(self, operand_node: object) -> APLArray:
        """Monadic ⎕FMT — handles both regular operands and FmtArgs."""
        from marple.formatting import format_num
        from marple.nodes import FmtArgs
        if isinstance(operand_node, FmtArgs):
            values = [self.evaluate(arg) for arg in operand_node.args]
            parts = [self._fmt_value(v) for v in values]
            all_chars: list[object] = []
            for p in parts:
                all_chars.extend(p.data)
                all_chars.append(" ")
            if all_chars:
                all_chars.pop()
            return APLArray([len(all_chars)], all_chars)
        operand = self.evaluate(operand_node)
        return self._fmt_value(operand)

    def _fmt_value(self, operand: APLArray) -> APLArray:
        """Format a single value as a character vector."""
        from marple.formatting import format_num
        if operand.shape == []:
            text = format_num(operand.data[0])
        elif len(operand.shape) == 1:
            if len(operand.data) > 0 and isinstance(operand.data[0], str):
                text = "".join(str(c) for c in operand.data)
            else:
                text = " ".join(format_num(x) for x in operand.data)
        else:
            text = str(operand)
        return APLArray([len(text)], list(text))

    def _sys_fmt_dyadic(self, left_node: object, right_node: object) -> APLArray:
        """Dyadic ⎕FMT — format with specification string."""
        from marple.fmt import dyadic_fmt
        from marple.nodes import FmtArgs
        left = self.evaluate(left_node)
        fmt_str = _apl_chars_to_str(left.data)
        if isinstance(right_node, FmtArgs):
            values = [self.evaluate(arg) for arg in right_node.args]
        else:
            right = self.evaluate(right_node)
            values = [right]
        return dyadic_fmt(fmt_str, values)

    def _sys_cr(self, operand: APLArray) -> APLArray:
        fn_name = _apl_chars_to_str(operand.data)
        source = self.env.get_source(fn_name)
        if source is None:
            raise DomainError("Not a defined function: " + fn_name)
        if isinstance(source, list):
            lines = source
        elif "\n" in source:
            lines = source.split("\n")
        else:
            lines = [source]
        max_len = max(len(l) for l in lines) if lines else 0
        flat: list[object] = []
        for line in lines:
            flat.extend(list(_ljust(line, max_len)))
        return APLArray([len(lines), max_len], flat)

    def _sys_fx(self, operand: APLArray) -> APLArray:
        from marple.dfn_binding import DfnBinding
        from marple.errors import APLError
        from marple.parser import parse
        if len(operand.shape) == 2:
            rows, cols = operand.shape
            lines = []
            for r in range(rows):
                row_chars = operand.data[r * cols : (r + 1) * cols]
                lines.append(_apl_chars_to_str(row_chars).rstrip())
            text = "\n".join(lines)
        else:
            text = _apl_chars_to_str(operand.data)
        parts = text.split("←", 1)
        if len(parts) < 2:
            raise DomainError("⎕FX requires an assignment: name←{body}")
        fn_name = parts[0].strip()
        from marple.executor import _newlines_to_diamonds
        source = _newlines_to_diamonds(text)
        tree = parse(source, self.env.class_dict())
        try:
            self.evaluate(tree)
        except APLError:
            raise DomainError("⎕FX: invalid function definition")
        val = self.env.get(fn_name)
        if not isinstance(val, DfnBinding):
            raise DomainError("⎕FX did not produce a function")
        self.env.set_source(fn_name, text.strip())
        if "⍺⍺" in text or "⍵⍵" in text:
            self.env.classify(fn_name, NC_OPERATOR)
            self.env.set_operator_arity(fn_name, 2 if "⍵⍵" in text else 1)
        chars = list(fn_name)
        return APLArray([len(chars)], chars)

    def _sys_dl(self, operand: APLArray) -> APLArray:
        import time as _time
        secs = float(operand.data[0])
        t0 = _time.time()
        _time.sleep(secs)
        return S(_time.time() - t0)

    def _sys_csv(self, operand: APLArray) -> APLArray:
        import csv as _csv
        import io as _io
        path = _apl_chars_to_str(operand.data)
        text = self.fs.read_text(path)
        reader = _csv.reader(_io.StringIO(text))
        headers = next(reader)
        col_names = []
        for h in headers:
            name = h.strip().replace(" ", "_")
            name = "".join(c if c.isalnum() or c == "_" else "_" for c in name)
            col_names.append(name)
        columns: list[list[str]] = [[] for _ in col_names]
        row_count = 0
        for row in reader:
            row_count += 1
            for i, val in enumerate(row):
                if i < len(columns):
                    columns[i].append(val.strip())
        for col_name, col_data in zip(col_names, columns):
            try:
                nums: list[int | float] = []
                for v in col_data:
                    if "." in v:
                        nums.append(float(v))
                    else:
                        nums.append(int(v))
                self.env.bind_name(col_name, APLArray([len(nums)], nums), NC_ARRAY)
            except (ValueError, TypeError):
                max_len = max((len(v) for v in col_data), default=0)
                chars: list[object] = []
                for v in col_data:
                    chars.extend(list(v.ljust(max_len)))
                self.env.bind_name(
                    col_name,
                    APLArray([len(col_data), max_len], chars),
                    NC_ARRAY,
                )
        return S(row_count)

    def _sys_nread(self, operand: APLArray) -> APLArray:
        path = _apl_chars_to_str(operand.data)
        text = self.fs.read_text(path)
        chars = list(text)
        return APLArray([len(chars)], chars) if chars else APLArray([0], [])

    def _sys_nexists(self, operand: APLArray) -> APLArray:
        path = _apl_chars_to_str(operand.data)
        return S(1 if self.fs.exists(path) else 0)

    def _sys_ndelete(self, operand: APLArray) -> APLArray:
        path = _apl_chars_to_str(operand.data)
        try:
            self.fs.delete(path)
        except OSError:
            raise DomainError("File not found: " + path)
        return S(0)

    def _sys_nwrite(self, left: APLArray, right: APLArray) -> APLArray:
        path = _apl_chars_to_str(right.data)
        text = _apl_chars_to_str(left.data)
        self.fs.write_text(path, text)
        return APLArray([0], [])



"""BufferedConsole — captures output for retrieval by PRIDE, Jupyter, etc."""

from marple.ports.console import Console


class BufferedConsole(Console):
    """Console adapter that buffers output for programmatic retrieval."""

    def __init__(self) -> None:
        self._output: list[str] = []

    def read_line(self, prompt: str) -> str | None:
        return None

    def write(self, text: str) -> None:
        self._output.append(text)

    def writeln(self, text: str) -> None:
        self._output.append(text + "\n")

    def clear(self) -> None:
        """Clear the output buffer."""
        self._output.clear()

    @property
    def output(self) -> str:
        """Return all captured output as a single string."""
        return "".join(self._output)

    @property
    def output_lines(self) -> list[str]:
        """Return captured output split into lines (no trailing empties)."""
        text = self.output
        if text.endswith("\n"):
            text = text[:-1]
        return text.split("\n") if text else []




import os
from datetime import datetime
try:
    from typing import Any, Callable
except ImportError:
    pass

from marple.arraymodel import APLArray, S
from marple.backend import to_list
from marple.ports.filesystem import FileSystem


def _is_dfn_binding(value: object) -> bool:
    """Check if value is any kind of dfn binding (old or new)."""
    return hasattr(value, 'dfn') and hasattr(value, 'env')


def _format_value(value: object) -> str | None:
    """Convert an APLArray value to APL source text."""
    if _is_dfn_binding(value):
        return None
    if not isinstance(value, APLArray):
        return None
    if value.is_scalar():
        v = value.data[0]
        if isinstance(v, str):
            return f"'{v}'"
        if isinstance(v, (int, float)) and v < 0:
            return f"¯{abs(v)}"
        return str(v)
    data = to_list(value.data)
    is_char = len(data) > 0 and all(isinstance(x, str) for x in data)
    if is_char:
        char_str = "".join(str(x) for x in data)
        quoted = f"'{char_str}'"
        if len(value.shape) == 1:
            return quoted
        shape_str = " ".join(str(s) for s in value.shape)
        return f"{shape_str}⍴{quoted}"

    def _fmt_num(x: object) -> str:
        if isinstance(x, (int, float)) and x < 0:
            return f"¯{abs(x)}"
        return str(x)

    data_str = " ".join(_fmt_num(x) for x in data)
    if len(value.shape) == 1:
        return data_str
    shape_str = " ".join(str(s) for s in value.shape)
    return f"{shape_str}⍴{data_str}"


def _sysvar_filename(name: str) -> str:
    """Convert system variable name to filesystem-safe filename.
    ⎕IO → __IO.apl"""
    return f"__{name[1:]}.apl"


def _entity_filename(name: str) -> str:
    """Convert entity name to filename."""
    if name.startswith("⎕"):
        return _sysvar_filename(name)
    return f"{name}.apl"


def _default_fs() -> FileSystem:
    from marple.adapters.os_filesystem import OsFileSystem
    return OsFileSystem()


def save_workspace(env: dict[str, Any], ws_dir: str,
                   fs: FileSystem | None = None) -> None:
    """Save workspace to a directory with one file per entity."""
    if fs is None:
        fs = _default_fs()
    fs.makedirs(ws_dir)

    # Write .ws marker
    wsid_val = env.get("⎕WSID", env.get("__wsid__", "CLEAR WS"))
    if isinstance(wsid_val, APLArray):
        wsid = "".join(str(c) for c in wsid_val.data)
    else:
        wsid = str(wsid_val)
    fs.write_text(os.path.join(ws_dir, ".ws"),
                  f"{wsid}\n{datetime.now().isoformat()}\n")

    # Track which files we write so we can clean up stale ones
    written_files: set[str] = {".ws"}
    sources: dict[str, str] = env.get("__sources__", {})

    # System variables that should not be saved (constants or managed separately)
    _SKIP_QUADS = {"⎕A", "⎕D", "⎕TS", "⎕EN", "⎕DM", "⎕WSID"}

    # Save system variables first
    for name in sorted(env):
        if name.startswith("⎕"):
            if name in _SKIP_QUADS:
                continue
            value = env[name]
            if isinstance(value, APLArray):
                formatted = _format_value(value)
                if formatted is not None:
                    filename = _entity_filename(name)
                    written_files.add(filename)
                    fs.write_text(os.path.join(ws_dir, filename),
                                  f"{name}←{formatted}\n")

    # Save user definitions
    for name in sorted(env):
        if name.startswith("⎕") or name.startswith("__"):
            continue
        if name in ("⍵", "⍺", "∇"):
            continue
        value = env[name]
        filename = _entity_filename(name)
        if _is_dfn_binding(value) and name in sources:
            written_files.add(filename)
            fs.write_text(os.path.join(ws_dir, filename),
                          f"{sources[name]}\n")
        elif isinstance(value, APLArray):
            formatted = _format_value(value)
            if formatted is not None:
                written_files.add(filename)
                fs.write_text(os.path.join(ws_dir, filename),
                              f"{name}←{formatted}\n")

    # Remove stale .apl files
    for existing in fs.listdir(ws_dir):
        if existing.endswith(".apl") and existing not in written_files:
            fs.delete(os.path.join(ws_dir, existing))


def load_workspace(env: Any, ws_dir: str,
                    evaluate: Callable[[str], Any] | None = None,
                    fs: FileSystem | None = None) -> None:
    """Load workspace from a directory.

    If evaluate is provided, it's called to execute each line.
    """
    if fs is None:
        fs = _default_fs()
    # Read .ws marker for WSID
    ws_file = os.path.join(ws_dir, ".ws")
    if fs.is_file(ws_file):
        text = fs.read_text(ws_file)
        wsid = text.split("\n")[0].strip()
        env["__wsid__"] = wsid
        env["⎕WSID"] = APLArray([len(wsid)], list(wsid))

    # Collect .apl files, system vars first
    files = sorted(fs.listdir(ws_dir))
    sys_files = [f for f in files if f.startswith("__") and f.endswith(".apl")]
    user_files = [f for f in files if not f.startswith("__") and f.endswith(".apl")]

    if evaluate is None:
        raise ValueError("evaluate callable is required for load_workspace")

    for filename in sys_files + user_files:
        filepath = os.path.join(ws_dir, filename)
        line = fs.read_text(filepath).strip()
        if line:
            evaluate(line)

    # Execute latent expression if present
    lx = env.get("⎕LX") if hasattr(env, 'get') else env.get("⎕LX")
    if lx is not None and isinstance(lx, APLArray) and len(lx.data) > 0:
        lx_text = "".join(str(c) for c in lx.data)
        if lx_text.strip():
            result = evaluate(lx_text)
            if isinstance(result, APLArray):
                return result
    return None


def list_workspaces(root: str, fs: FileSystem | None = None) -> list[str]:
    """List workspace names under the given root directory."""
    if fs is None:
        fs = _default_fs()
    if not fs.is_dir(root):
        return []
    result = []
    for name in sorted(fs.listdir(root)):
        ws_dir = os.path.join(root, name)
        if fs.is_dir(ws_dir) and fs.is_file(os.path.join(ws_dir, ".ws")):
            result.append(name)
    return result




try:
    from typing import Any
except ImportError:
    pass

import sys

from marple.engine import Interpreter
from marple.errors import APLError
from marple.formatting import format_result  # re-export for backward compat
from marple.ports.console import Console
from marple.system_commands import run_system_command


def run_repl(interp: Interpreter, console: Console) -> None:
    """Run the REPL loop using the given Console for I/O."""
    from marple import __version__ as ver
    console.writeln(f"MARPLE v{ver} - Mini APL in Python")
    console.writeln("CLEAR WS")
    console.writeln("")
    while True:
        line = console.read_line("      ")
        if line is None:
            break
        line = line.strip()
        if not line:
            continue
        if line.startswith(")"):
            output, should_exit = run_system_command(interp, line)
            if output:
                console.writeln(output)
            if should_exit:
                break
            continue
        try:
            r = interp.execute(line)
            if not r.silent:
                console.writeln(r.display_text)
        except APLError as e:
            console.writeln(str(e))
        except Exception as e:
            console.writeln(f"ERROR: {e}")


def main() -> None:
    if len(sys.argv) > 1:
        from marple.script import run_script
        path = sys.argv[1]
        for line in run_script(path):
            print(line)
        return

    from marple.adapters.terminal_console import TerminalConsole
    console = TerminalConsole()
    interp = Interpreter(console=console)
    run_repl(interp, console)


if __name__ == "__main__":
    main()



"""OsFileSystem — real FileSystem adapter using os and open()."""

import os

from marple.ports.filesystem import FileSystem


class OsFileSystem(FileSystem):
    """FileSystem adapter that uses the real OS filesystem."""

    def read_text(self, path: str) -> str:
        with open(path) as f:
            return f.read()

    def write_text(self, path: str, content: str) -> None:
        with open(path, "w") as f:
            f.write(content)

    def exists(self, path: str) -> bool:
        try:
            os.stat(path)
            return True
        except OSError:
            return False

    def is_file(self, path: str) -> bool:
        try:
            return (os.stat(path)[0] & 0x8000) != 0
        except OSError:
            return False

    def is_dir(self, path: str) -> bool:
        try:
            return (os.stat(path)[0] & 0x4000) != 0
        except OSError:
            return False

    def delete(self, path: str) -> None:
        os.remove(path)

    def makedirs(self, path: str) -> None:
        parts = path.replace("\\", "/").split("/")
        current = ""
        for part in parts:
            if not part:
                current = "/"
                continue
            current = current + part if current.endswith("/") else current + "/" + part
            try:
                os.mkdir(current)
            except OSError:
                pass

    def listdir(self, path: str) -> list[str]:
        return os.listdir(path)




try:
    from typing import Any
except ImportError:
    pass

from marple.adapters.buffered_console import BufferedConsole
from marple.engine import Interpreter
from marple.errors import APLError
from marple.ports.filesystem import FileSystem


PROMPT = "      "


def run_script(path: str, fs: FileSystem | None = None) -> list[str]:
    """Run an APL script file line by line.

    Echoes each input line with the REPL prompt, followed by output.
    Stops on first error with an error message including line number.
    """
    if fs is None:
        from marple.adapters.os_filesystem import OsFileSystem
        fs = OsFileSystem()
    console = BufferedConsole()
    interp = Interpreter(fs=fs, console=console)
    output: list[str] = []
    text = fs.read_text(path)
    accum = ""
    start_lineno = 1
    for lineno, line in enumerate(text.splitlines(), 1):
        line = line.strip()
        if not line:
            continue
        output.append(f"{PROMPT}{line}")
        if line.startswith("⍝") and not accum:
            continue
        if line.startswith(")") and not accum:
            from marple.system_commands import run_system_command
            cmd_output, _ = run_system_command(interp, line)
            if cmd_output:
                output.append(cmd_output)
            continue
        if accum:
            accum += "\n" + line
        else:
            accum = line
            start_lineno = lineno
        # Check if braces are balanced
        if accum.count("{") > accum.count("}"):
            continue
        try:
            console.clear()
            r = interp.execute(accum)
            console_output = console.output
            if console_output:
                output.append(console_output.rstrip("\n"))
            if not r.silent:
                output.append(r.display_text)
        except APLError as e:
            output.append(f"{e} at line {start_lineno}")
            output.append(f"  {accum}")
            break
        except Exception as e:
            output.append(f"ERROR at line {start_lineno}: {e}")
            output.append(f"  {accum}")
            break
        accum = ""
    return output



"""Convert APLArray to HTML for Jupyter rich display."""

import html as _html

from marple.arraymodel import APLArray
from marple.formatting import format_num

ARRAY_CSS = """<style>
.apl-array {
    border-collapse: collapse;
    font-family: 'APL385 Unicode', 'DejaVu Sans Mono', monospace;
    margin: 4px 0;
}
.apl-array td {
    border: 1px solid #ddd;
    padding: 2px 8px;
    text-align: right;
    white-space: nowrap;
}
.apl-array .char-cell { text-align: left; }
.apl-array tr:nth-child(even) { background: #f7f7f7; }
.apl-slice {
    margin: 6px 0;
    padding-left: 10px;
    border-left: 3px solid #4a86c8;
}
.apl-slice-label {
    color: #888;
    font-size: 0.85em;
    font-family: 'APL385 Unicode', 'DejaVu Sans Mono', monospace;
}
</style>"""


def _is_char_data(data: object) -> bool:
    return hasattr(data, '__len__') and len(data) > 0 and all(isinstance(x, str) for x in data)  # type: ignore[arg-type]


def _cell_html(value: object, is_char: bool) -> str:
    if isinstance(value, str):
        return f'<td class="char-cell">{_html.escape(value)}</td>'
    return f'<td>{_html.escape(format_num(value))}</td>'


def aplarray_to_html(arr: APLArray) -> str:
    """Convert an APLArray to an HTML representation."""
    is_char = _is_char_data(arr.data)

    # Scalar
    if arr.shape == []:
        val = arr.data[0]
        text = _html.escape(str(val) if isinstance(val, str) else format_num(val))
        return f'<span class="apl-scalar">{text}</span>'

    # Vector
    if len(arr.shape) == 1:
        if is_char:
            return f'<span class="apl-scalar">{_html.escape("".join(str(c) for c in arr.data))}</span>'
        if len(arr.data) == 0:
            return '<table class="apl-array"><tr></tr></table>'
        cells = ''.join(_cell_html(v, is_char) for v in arr.data)
        return f'<table class="apl-array"><tr>{cells}</tr></table>'

    # Matrix
    if len(arr.shape) == 2:
        return _matrix_html(arr, is_char)

    # Rank 3+
    return _high_rank_html(arr, is_char)


def _matrix_html(arr: APLArray, is_char: bool) -> str:
    rows, cols = arr.shape
    html_rows = []
    for r in range(rows):
        cells = ''.join(
            _cell_html(arr.data[r * cols + c], is_char) for c in range(cols))
        html_rows.append(f'<tr>{cells}</tr>')
    return f'<table class="apl-array">{"".join(html_rows)}</table>'


def _high_rank_html(arr: APLArray, is_char: bool) -> str:
    outer = arr.shape[0]
    inner_shape = arr.shape[1:]
    slice_size = 1
    for s in inner_shape:
        slice_size *= s
    slices = []
    for i in range(outer):
        sub = APLArray(list(inner_shape),
                       list(arr.data[i * slice_size:(i + 1) * slice_size]))
        inner_html = aplarray_to_html(sub)
        slices.append(
            f'<div class="apl-slice">'
            f'<span class="apl-slice-label">[{i + 1}]</span>'
            f'{inner_html}</div>')
    return ''.join(slices)



"""Dyadic primitive function dispatch for MARPLE."""

from __future__ import annotations

from typing import Any, TYPE_CHECKING

from marple.arraymodel import APLArray, S

if TYPE_CHECKING:
    from marple.environment import Environment
import random as _random

from marple.errors import DomainError, LengthError
from marple.functions import (
    add,
    binomial,
    subtract,
    multiply,
    divide,
    maximum,
    minimum,
    power,
    logarithm,
    residue,
    logical_and,
    logical_or,
    circular,
    less_than,
    less_equal,
    equal,
    greater_equal,
    greater_than,
    not_equal,
)
from marple.structural import (
    catenate,
    drop,
    encode,
    decode,
    expand,
    from_array,
    index_of,
    membership,
    replicate,
    replicate_first,
    reshape,
    rotate,
    rotate_first,
    take,
    matrix_divide,
)


class DyadicFunctionBinding:
    """Dispatches and applies dyadic primitive functions."""

    _SIMPLE: dict[str, object] = {
        "+": add,
        "-": subtract,
        "×": multiply,
        "÷": divide,
        "⌈": maximum,
        "⌊": minimum,
        "*": power,
        "⍟": logarithm,
        "|": residue,
        "∧": logical_and,
        "∨": logical_or,
        "⍴": reshape,
        ",": catenate,
        "↑": take,
        "↓": drop,
        "⌽": rotate,
        "⊖": rotate_first,
        "⊤": encode,
        "⊥": decode,
        "/": replicate,
        "⌿": replicate_first,
        "\\": expand,
        "⌹": matrix_divide,
        "○": circular,
        "!": binomial,
        "≡": lambda a, o: S(1 if a == o else 0),
        "≢": lambda a, o: S(0 if a == o else 1),
    }

    _ENV_DEPENDENT: dict[str, str] = {
        "<": "_less_than",
        "≤": "_less_equal",
        "=": "_equal",
        "≥": "_greater_equal",
        ">": "_greater_than",
        "≠": "_not_equal",
        "⍳": "_index_of",
        "∈": "_membership",
        "⌷": "_from",
        "⍕": "_format",
        "?": "_deal",
    }

    def __init__(self, env: Environment) -> None:
        self._env = env

    def apply(self, glyph: str, left: APLArray, right: APLArray) -> APLArray:
        """Apply a dyadic primitive function to left and right arguments."""
        method_name = self._ENV_DEPENDENT.get(glyph)
        if method_name is not None:
            return getattr(self, method_name)(left, right)
        func = self._SIMPLE.get(glyph)
        if func is not None:
            return func(left, right)  # type: ignore[operator]
        raise DomainError(f"Unknown dyadic function: {glyph}")

    def _less_than(self, left: APLArray, right: APLArray) -> APLArray:
        return less_than(left, right, self._env.ct)

    def _less_equal(self, left: APLArray, right: APLArray) -> APLArray:
        return less_equal(left, right, self._env.ct)

    def _equal(self, left: APLArray, right: APLArray) -> APLArray:
        return equal(left, right, self._env.ct)

    def _greater_equal(self, left: APLArray, right: APLArray) -> APLArray:
        return greater_equal(left, right, self._env.ct)

    def _greater_than(self, left: APLArray, right: APLArray) -> APLArray:
        return greater_than(left, right, self._env.ct)

    def _not_equal(self, left: APLArray, right: APLArray) -> APLArray:
        return not_equal(left, right, self._env.ct)

    def _index_of(self, left: APLArray, right: APLArray) -> APLArray:
        return index_of(left, right, self._env.io, self._env.ct)

    def _membership(self, left: APLArray, right: APLArray) -> APLArray:
        return membership(left, right, self._env.ct)

    def _from(self, left: APLArray, right: APLArray) -> APLArray:
        return from_array(left, right, self._env.io)

    def _format(self, left: APLArray, right: APLArray) -> APLArray:
        if left.is_scalar():
            width = int(left.data[0])
            precision = None
        else:
            width = int(left.data[0])
            precision = int(left.data[1]) if len(left.data) > 1 else None
        values = right.data if not right.is_scalar() else [right.data[0]]
        result_chars: list[str] = []
        for v in values:
            if precision is not None:
                formatted = f"{float(v):.{precision}f}"
            else:
                formatted = str(v)
            padded = " " * max(0, width - len(formatted)) + formatted
            result_chars.extend(list(padded))
        return APLArray([len(result_chars)], result_chars)

    def _deal(self, left: APLArray, right: APLArray) -> APLArray:
        io = self._env.io
        n = int(left.data[0])
        m = int(right.data[0])
        if n > m:
            raise LengthError(f"Deal: cannot choose {n} from {m}")
        result = _random.sample(range(io, m + io), n)
        return APLArray([n], result)

    # Comparison functions for operator use (reduce/scan)
    # ct defaults to 0, so they work with 2 args
    _OPERATOR_COMPARISONS: dict[str, object] = {
        "<": less_than,
        "≤": less_equal,
        "=": equal,
        "≥": greater_equal,
        ">": greater_than,
        "≠": not_equal,
    }

    @classmethod
    def resolve(cls, glyph: str) -> object:
        """Return the callable for a glyph, for use by operators."""
        func = cls._SIMPLE.get(glyph)
        if func is not None:
            return func
        func = cls._OPERATOR_COMPARISONS.get(glyph)
        if func is not None:
            return func
        raise DomainError(f"Unknown function for operator: {glyph}")

    def resolve_with_env(self, glyph: str) -> object:
        """Return a callable for a glyph, including env-dependent functions."""
        method_name = self._ENV_DEPENDENT.get(glyph)
        if method_name is not None:
            method = getattr(self, method_name)
            return lambda a, o, _m=method: _m(a, o)
        func = self._SIMPLE.get(glyph)
        if func is not None:
            return func
        raise DomainError(f"Unknown function for operator: {glyph}")



"""System command dispatcher for MARPLE.

Returns strings instead of printing — usable by REPL, web server, and Jupyter.
"""

import os

from marple.arraymodel import APLArray
from marple.engine import Interpreter


def run_system_command(interp: Interpreter, line: str) -> tuple[str, bool]:
    """Execute a system command. Returns (output_text, should_exit)."""
    cmd = line[1:].split()[0].lower() if line[1:].strip() else ""
    handler = _COMMANDS.get(cmd)
    if handler is not None:
        return handler(interp, line)
    return f"Unknown command: {line}", False


def _cmd_off(interp: Interpreter, line: str) -> tuple[str, bool]:
    return "", True


def _cmd_clear(interp: Interpreter, line: str) -> tuple[str, bool]:
    from marple.environment import Environment
    interp.env = Environment(io=1)
    return "CLEAR WS", False


def _cmd_wsid(interp: Interpreter, line: str) -> tuple[str, bool]:
    parts = line.split(None, 1)
    if len(parts) > 1:
        name = parts[1].strip()
        interp.env["⎕WSID"] = APLArray([len(name)], list(name))
        return name, False
    return "".join(str(c) for c in interp.env["⎕WSID"].data), False


def _cmd_ops(interp: Interpreter, line: str) -> tuple[str, bool]:
    names = [name for name, _ in interp.env.list_operators()]
    return "  ".join(names), False


def _cmd_fns(interp: Interpreter, line: str) -> tuple[str, bool]:
    parts = line.split(None, 1)
    if len(parts) > 1:
        ns_name = parts[1].strip()
        if ns_name.startswith("$"):
            from marple.namespace import load_system_workspace
            import marple.stdlib
            f = marple.stdlib.__file__
            stdlib_path = f[:f.rfind("/")] if "/" in f else f[:f.rfind("\\")]
            sys_ws = load_system_workspace(stdlib_path)
            ns_parts = ns_name.split("::")[1:] if "::" in ns_name else []
            ns = sys_ws.resolve(ns_parts) if ns_parts else sys_ws
            if ns is not None and hasattr(ns, "list_names"):
                return "  ".join(ns.list_names()), False
            return f"Namespace not found: {ns_name}", False
        return f"Namespace not found: {ns_name}", False
    names = [name for name, _ in interp.env.list_functions()]
    return "  ".join(names), False


def _cmd_vars(interp: Interpreter, line: str) -> tuple[str, bool]:
    names = [name for name, _ in interp.env.list_variables()]
    return "  ".join(names), False


def _cmd_lib(interp: Interpreter, line: str) -> tuple[str, bool]:
    from marple.workspace import list_workspaces
    from marple.config import get_workspaces_dir
    ws_root = get_workspaces_dir()
    workspaces = list_workspaces(ws_root)
    return "  ".join(workspaces) if workspaces else "(none)", False


def _cmd_save(interp: Interpreter, line: str) -> tuple[str, bool]:
    from marple.workspace import save_workspace
    from marple.config import get_workspaces_dir
    parts = line.split(None, 1)
    if len(parts) > 1:
        name = parts[1].strip()
        interp.env["⎕WSID"] = APLArray([len(name)], list(name))
    wsid = "".join(str(c) for c in interp.env["⎕WSID"].data)
    if wsid == "CLEAR WS":
        return "ERROR: No workspace ID set. Use )WSID name first.", False
    ws_root = get_workspaces_dir()
    env_dict: dict[str, object] = {}
    for name in interp.env.quad_var_names():
        env_dict[name] = interp.env[name]
    for name in interp.env.user_names():
        env_dict[name] = interp.env[name]
    env_dict["__sources__"] = interp.env.sources()
    env_dict["__wsid__"] = wsid
    try:
        save_workspace(env_dict, os.path.join(ws_root, wsid))
        return f"{wsid} SAVED", False
    except Exception as e:
        return f"ERROR: {e}", False


def _cmd_load(interp: Interpreter, line: str) -> tuple[str, bool]:
    from marple.workspace import load_workspace
    from marple.config import get_workspaces_dir
    parts = line.split(None, 1)
    if len(parts) < 2:
        return "ERROR: )LOAD requires a workspace name", False
    name = parts[1].strip()
    ws_root = get_workspaces_dir()
    ws_dir = os.path.join(ws_root, name)
    if not os.path.isdir(ws_dir):
        return f"ERROR: Workspace not found: {name}", False
    from marple.environment import Environment
    interp.env = Environment(io=1)
    try:
        from marple.formatting import format_result
        lx_result = load_workspace(interp.env, ws_dir, evaluate=interp.run)
        wsid = "".join(str(c) for c in interp.env["⎕WSID"].data)
        output = wsid
        if lx_result is not None:
            output += "\n" + format_result(lx_result, interp.env)
        return output, False
    except Exception as e:
        return f"ERROR: {e}", False


from typing import Callable

_CmdFn = Callable[[Interpreter, str], tuple[str, bool]]

_COMMANDS: dict[str, _CmdFn] = {
    "off": _cmd_off,
    "clear": _cmd_clear,
    "wsid": _cmd_wsid,
    "fns": _cmd_fns,
    "ops": _cmd_ops,
    "vars": _cmd_vars,
    "lib": _cmd_lib,
    "save": _cmd_save,
    "load": _cmd_load,
}
